from typing import List
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction
from etk.etk_exceptions import ExtractorError


class DecodingValueExtractor(Extractor):
    """
    **Description**
        This class takes a 'decoding_dict' as reference, decoding the input text based on the 'decoding_dict'

    Examples:
        ::

            decoding_dict = {
            'CA': 'California',
            'ny': 'New York',
            'AZ': ' Arizona',
            ' TX ': 'Texas',
            ' fl': 'Florida',
            }
            decoding_value_extractor = DecodingValueExtractor(decoding_dict=decoding_dict,
                                                            extractor_name='default_decoding',
                                                            case_sensitive=True,
                                                            strip_key=False,
                                                            strip_value=True)
            decoding_value_extractor.extract(value=value_to_be_decoded)

    """
    def __init__(self,
                 decoding_dict: dict,
                 extractor_name: str,
                 default_action: str='delete',
                 case_sensitive: bool=False,
                 strip_key: bool=True,
                 strip_value: bool=False,
                 ) -> None:
        """

        Args:
            decoding_dict: dict -> a python dictionary for decoding values
            extractor_name: str -> extractor name
            default_action: enum['delete'] ->  what if the value not matched in dictionary
            case_sensitive: bool -> matching the key and value strictly or ignore cases
            strip_key: bool -> strip key and value for matching or not
            strip_value: bool -> return the striped value if matched or the original value
        """
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="dictionary",
                           name=extractor_name)
        if case_sensitive and not strip_key:
            self._decoding_dict = decoding_dict
        else:
            new_dict = {}
            if not strip_key:   # not case_sensitive, ignore cases
                for k in decoding_dict:
                    new_dict[k.lower()] = decoding_dict[k]
            elif case_sensitive:   # strip key
                for k in decoding_dict:
                    new_dict[k.strip()] = decoding_dict[k]
            else:   # ignore case AND strip key
                for k in decoding_dict:
                    new_dict[k.lower().strip()] = decoding_dict[k]
            self._decoding_dict = new_dict

        self._case_sensitive = case_sensitive
        self._default_action = default_action
        self._strip_key = strip_key
        self._strip_value = strip_value

        self._joiner = " "

    def extract(self, value: str) -> List[Extraction]:
        """

        Args:
            value (str): the value to be decode

        Returns:
            List[Extraction]: actually a single Extraction wrapped in a list if there is a match

        """

        to_match = value.lower() if not self._case_sensitive else value
        to_match = to_match.strip() if self._strip_key else to_match

        if to_match in self._decoding_dict:
            extraction = self._wrap_result(self._decoding_dict[to_match], value)
            return [extraction] if extraction else list()
        else:
            if self._default_action == 'delete':
                return list()

        return list()

    def _wrap_result(self, value: str, original_key: str) -> Extraction or None:
        """

        Args:
            value: the decoded value
            original_key: the original string value to be decode

        Returns: an Extraction if everything goes well

        """
        try:
            value = value.strip() if self._strip_value else value
            e = Extraction(value, self.name, start_char=0, end_char=len(str(value)))
            return e
        except Exception as e:
            print('fail to wrap dictionary extraction: ', original_key, value)
            raise ExtractorError('Exception: ' + str(e))



from etk.extractors.regex_extractor import RegexExtractor
from etk.extractor import Extractor, InputType
from typing import List
from etk.extraction import Extraction


class CryptographicHashExtractor(Extractor):
    """
    **Description**
        This class inherits RegexExtractor by predefine the regex patterns for md5, sha1 and sha256

    Examples:
        ::

            cryptographic_hash_extractor = CryptographicHashExtractor()
            cryptographic_hash_extractor.extract(text=input_doc)

    """
    def __init__(self):
        e_name = 'cryptographic hash extractor'
        self._regex_extractors = [
            RegexExtractor(r"(\b[a-fA-F\d]{32}\b)", 'md5 '+e_name, general_tag='md5'),
            RegexExtractor(r"(\b[0-9a-f]{40}\b)", 'sha1 '+e_name, general_tag='sha1'),
            RegexExtractor(r"(\b[A-Fa-f0-9]{64}\b)", 'sha256 '+e_name, general_tag='sha256'),
        ]
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="regex",
                           name=e_name)

    @property
    def regex_extractors(self):
        return self._regex_extractors

    def extract(self, text: str) -> List[Extraction]:
        """

        Args:
            text (str): The input source to be processed

        Returns:
            List[Extraction]: The list of extractions returned by CryptographicHashExtractor

        """

        res = list()
        for e in self.regex_extractors:
            res = res+e.extract(text)
        return res



# import unittest
# from etk.extractors.dbpedia_spotlight_extractor import DBpediaSpotlightExtractor
#
#
# class TestDBPediaSpotlightExtractor(unittest.TestCase):
#
#     def test_DBPediaSpotlight_extractor(self) -> None:
#         test_get_attr = False
#         filters = ['Person', 'Place', 'Organisation']
#         confidence = 0.5
#         extractor = DBpediaSpotlightExtractor(extractor_name='dbPedia_extractor',
#                                               search_url='http://model.dbpedia-spotlight.org/en/annotate',
#                                               get_attr=test_get_attr,
#                                               get_attr_url="http://dbpedia.org/sparql")
#         text = 'Cohen has not been charged with a crime, but the FBI raid of his home, \
#                 hotel room and office in early April revealed that prosecutors had zeroed in on \
#                 his personal financial dealings, including the payment he made to porn star \
#                 Stormy Daniels on Trump\'s behalf before the election.'
#         extracted = list()
#         results = extractor.extract(text, confidence=confidence, filter=filters)
#
#         # if not test_get_attr:
#
#         for i in results:
#             extracted_value = {
#                 'value': i.value,
#                 'confidence': i.confidence,
#                 'start_char': i.provenance['start_char'],
#                 'end_char': i.provenance['end_char'],
#             }
#             extracted.append(extracted_value)
#         expected = \
#             [
#                 {"value":
#                      {"surface_form": "FBI",
#                       "uri": "http://dbpedia.org/resource/Federal_Bureau_of_Investigation",
#                       "types":
#                           ["Wikidata:Q43229",
#                            "Wikidata:Q327333",
#                            "Wikidata:Q24229398",
#                            "DUL:SocialPerson",
#                            "DUL:Agent",
#                            "Schema:Organization",
#                            "DBpedia:Organisation",
#                            "DBpedia:GovernmentAgency",
#                            "DBpedia:Agent"],
#                       "similarity_scores": 0.9999999999998863},
#                  "confidence": 0.5,
#                  "start_char": 49,
#                  "end_char": 52},
#
#                 {"value":
#                      {"surface_form": "financial dealings",
#                       "uri": "http://dbpedia.org/resource/Yasser_Arafat",
#                       "types":
#                           ["Http://xmlns.com/foaf/0.1/Person",
#                            "Wikidata:Q82955",
#                            "Wikidata:Q5",
#                            "Wikidata:Q30461",
#                            "Wikidata:Q24229398",
#                            "Wikidata:Q215627",
#                            "DUL:NaturalPerson",
#                            "DUL:Agent",
#                            "Schema:Person",
#                            "DBpedia:President",
#                            "DBpedia:Politician",
#                            "DBpedia:Person",
#                            "DBpedia:Agent"],
#                       "similarity_scores": 0.9999999999648139},
#                  "confidence": 0.5,
#                  "start_char": 196,
#                  "end_char": 214},
#
#                 {"value":
#                      {"surface_form": "Stormy Daniels",
#                       "uri": "http://dbpedia.org/resource/Stormy_Daniels",
#                       "types":
#                           ["Http://xmlns.com/foaf/0.1/Person",
#                            "Wikidata:Q5",
#                            "Wikidata:Q488111",
#                            "Wikidata:Q483501",
#                            "Wikidata:Q33999",
#                            "Wikidata:Q24229398",
#                            "Wikidata:Q215627",
#                            "DUL:NaturalPerson",
#                            "DUL:Agent",
#                            "Schema:Person",
#                            "DBpedia:Person",
#                            "DBpedia:Artist",
#                            "DBpedia:Agent",
#                            "DBpedia:AdultActor",
#                            "DBpedia:Actor"],
#                       "similarity_scores": 1.0},
#                  "confidence": 0.5,
#                  "start_char": 275,
#                  "end_char": 289},
#
#                 {"value":
#                      {"surface_form": "Trump",
#                       "uri": "http://dbpedia.org/resource/Donald_Trump",
#                       "types":
#                           ["Http://xmlns.com/foaf/0.1/Person",
#                            "Wikidata:Q5",
#                            "Wikidata:Q24229398",
#                            "Wikidata:Q215627",
#                            "DUL:NaturalPerson",
#                            "DUL:Agent",
#                            "Schema:Person",
#                            "DBpedia:Person",
#                            "DBpedia:Agent"],
#                       "similarity_scores": 0.9888612234881796},
#                  "confidence": 0.5,
#                  "start_char": 293,
#                  "end_char": 298}]
#
#         result_count = 0
#         while result_count < len(extracted):
#             exact_form = extracted[result_count]['value']['surface_form']
#             index_form = text[extracted[result_count]['start_char']:extracted[result_count]['end_char']]
#             self.assertEqual(exact_form, index_form)
#             self.assertEqual(extracted[result_count], expected[result_count])
#             result_count += 1
#
#         # else:
#         test_get_attr = True
#         extractor = DBpediaSpotlightExtractor(extractor_name='dbPedia_extractor',
#                                               search_url='http://model.dbpedia-spotlight.org/en/annotate',
#                                               get_attr=test_get_attr,
#                                               get_attr_url="http://dbpedia.org/sparql")
#         extracted = list()
#         results = extractor.extract(text, confidence=confidence, filter=filters)
#         for i in results:
#             extracted_value = {
#                 'value': i.value,
#                 'confidence': i.confidence
#             }
#             extracted.append(extracted_value)
#         attr_len = [47, 41, 34, 55]
#         first_attr_cat = ['Federal Bureau of Investigation', 'Yasser Arafat', 'Stormy Daniels', 'Donald Trump']
#
#         result_count = 0
#         while result_count < len(extracted):
#             self.assertEqual(attr_len[result_count], len(extracted[result_count]['value']['attributes']))
#             self.assertEqual(first_attr_cat[result_count],
#                              extracted[result_count]['value']['attributes']['rdf-schema#label'][0])
#             result_count += 1
#
#
# if __name__ == '__main__':
#     unittest.main()



import unittest
from etk.extractors.bitcoin_address_extractor import BitcoinAddressExtractor
from etk.extractors.cryptographic_hash_extractor import CryptographicHashExtractor
from etk.extractors.cve_extractor import CVEExtractor
from etk.extractors.hostname_extractor import HostnameExtractor
from etk.extractors.ip_address_extractor import IPAddressExtractor
from etk.extractors.url_extractor import URLExtractor


class TestMultiDataExtractor(unittest.TestCase):

    def test_bitcoin_address_extractor(self) -> None:
        text = "1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2 3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLy " \
               "bc1qar0srrr7xfkvy5l643lydnw9re59gtzzwf5mdq bc1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2 " \
               "1AGNa15ZQXAZUgFiqJ3i7Z2DPU2J6hW62i 17NdbrSGoUotzeGCcMMCqnFkEvLymoou9j " \
               "bc1qc7slrfxkknqcq2jevvvkdgvrt8080852dfjewde450xdlk4ugp7szw5tk9 " \
               "bc1qar0srrr7xfkvy5l643lydnw9re59gtzzwf5mdq"
        e = BitcoinAddressExtractor()

        results = [x.value for x in e.extract(text)]
        expected = [
            "1BvBMSEYstWetqTFn5Au4m4GFg7xJaNVN2",
            "3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLy",
            "1AGNa15ZQXAZUgFiqJ3i7Z2DPU2J6hW62i",
            "17NdbrSGoUotzeGCcMMCqnFkEvLymoou9j"
        ]

        self.assertEqual(results, expected)

    def test_cryptographic_hash_extractor(self) -> None:
        text = "Sample 9e107d9d372bb6826bd81d3542a419d6 e4d909c290d0fb1ca068ffaddf22cbd0 " \
               "SHA1(The quick brown fox jumps over the lazy dog)\ngives hexadecimal: " \
               "2fd4e1c67a2d28fced849ee1bb76e7391b93eb12\ngives Base64 binary to ASCII text encoding: " \
               "L9ThxnotKPzthJ7hu3bnORuT6xI="
        e = CryptographicHashExtractor()

        results = [x.value for x in e.extract(text)]
        expected = [
            "9e107d9d372bb6826bd81d3542a419d6",
            "e4d909c290d0fb1ca068ffaddf22cbd0",
            "2fd4e1c67a2d28fced849ee1bb76e7391b93eb12"
        ]

        self.assertEqual(results, expected)

    def test_cve_extractor(self) -> None:
        text = "Sample cves are CVE-1993-1344,  CVE-2016-7654321, CVE-2006-1232 and CVE-1993-1344."
        e = CVEExtractor()

        results = [x.value for x in e.extract(text)]
        expected = [
            "CVE-1993-1344",
            "CVE-2016-7654321",
            "CVE-2006-1232",
            "CVE-1993-1344"
        ]

        self.assertEqual(results, expected)

    def test_hostname_extractor(self) -> None:
        text = "check sites brun1989.itldc-customer.net, www.google.com and " \
               "https://www.google.com go shopping with amazon.com"
        e = HostnameExtractor()

        results = [x.value for x in e.extract(text)]
        expected = [
            "brun1989.itldc-customer.net",
            "www.google.com",
            "www.google.com",
            "amazon.com"
        ]

        self.assertEqual(results, expected)

    def test_ip_address_extractor(self) -> None:
        text = "my ip address is 192.168.0.1 , 193.14.1.1 and 193.14.1.1 not 193.14"
        e = IPAddressExtractor()

        results = [x.value for x in e.extract(text)]
        expected = [
            "192.168.0.1",
            "193.14.1.1",
            "193.14.1.1"
        ]

        self.assertEqual(results, expected)

    def test_url_extractor(self) -> None:
        text = "https://foo_bar.example.com/ http://foo_bar.example.com/ https://foobar.example.com/xxx.php  " \
               "%%%###http://isi.edu/abx  dig.isi.edu/index.html"
        e = URLExtractor(True)

        results = [x.value for x in e.extract(text)]
        expected = [
            "https://foo_bar.example.com/",
            "http://foo_bar.example.com/",
            "https://foobar.example.com/xxx.php",
            "http://isi.edu/abx",
            "dig.isi.edu/index.html"
        ]

        self.assertEqual(results, expected)


if __name__ == '__main__':
    unittest.main()


from spacy.tokens import Token
from etk.tokenizer import Tokenizer
from typing import List, Any, Dict
import numbers
import datetime
import re


class ExtractableBase(object):
    """
    Encapsulates a value that can be used as an input to an Extractor
    """

    def __init__(self) -> None:
        self._value = None

    @property
    def value(self) -> Any:
        """
        Returns: Whatever value is returned by JSONPath
        """
        return self._value

    def get_string(self, joiner: str = " ") -> str:
        """
        Args:
            joiner(str): if the value of an extractable is not a string, join the elements
            using this string to separate them.

        Returns: the value of the segment as a string, using a default method to convert
        objects to strings.
        """
        if not self._value:
            return ""
        elif isinstance(self._value, list):
            return self.list2str(self._value, joiner)
        elif isinstance(self._value, dict):
            return self.dict2str(self._value, joiner)
        elif isinstance(self.value, numbers.Number):
            return str(self.value)
        elif isinstance(self._value, datetime.date):
            return self._value.strftime("%Y-%m-%d")
        elif isinstance(self._value, datetime.datetime):
            return self._value.isoformat()
        else:
            return self._value

    def list2str(self, l: List, joiner: str) -> str:
        """
        Convert list to str as input for tokenizer

        Args:
            l (list): list for converting
            joiner (str): join the elements using this string to separate them.

        Returns: the value of the list as a string

        """
        result = str()
        for item in l:
            if isinstance(item, list):
                result = result + self.list2str(item, joiner) + joiner
            elif isinstance(item, dict):
                result = result + self.dict2str(item, joiner) + joiner
            elif item:
                result = result + str(item) + joiner
        return result

    def dict2str(self, d: Dict, joiner: str) -> str:
        """
        Convert dict to str as input for tokenizer

        Args:
            d (dict): dict for converting
            joiner (str): join the elements using this string to separate them.

        Returns: the value of the dict as a string

        """
        result = str()
        for key in d:
            result = result + str(key) + " : "
            if isinstance(d[key], list):
                result = result + self.list2str(d[key], joiner) + joiner
            elif isinstance(d[key], dict):
                result = result + self.dict2str(d[key], joiner) + joiner
            elif d[key]:
                result = result + str(d[key]) + joiner
        return result


class Extractable(ExtractableBase):
    """
    A single extraction or a single segment
    """

    def __init__(self, value=None, prov_id=None) -> None:
        ExtractableBase.__init__(self)
        self.tokenize_results = dict()
        self._value = value
        self.prov_id = prov_id

    def get_tokens(self, tokenizer: Tokenizer) -> List[Token]:
        """
        Tokenize this Extractable.

        If the value is a string, it returns the tokenized version of the string. Else, convert to string with
        get_string method

        As it is common to need the same tokens for multiple extractors, the Extractable should cache the
        tokenize results, keyed by segment and tokenizer so that given the same segment and tokenizer,
        the same results are returned. If the same segment is given, but different tokenizer, the different
        results are cached separately.

        Args:
            tokenizer (Tokenizer)

        Returns: a sequence of tokens.
        """

        if (self, tokenizer) in self.tokenize_results:
            return self.tokenize_results[(self, tokenizer)]
        else:
            segment_value_for_tokenize = self.get_string()
            tokens = tokenizer.tokenize(segment_value_for_tokenize)
            self.tokenize_results[(self, tokenizer)] = tokens
            return tokens

    @property
    def prov_id(self):
        return self.__prov_id

    @prov_id.setter
    def prov_id(self, prov_id):
        self.__prov_id = prov_id


class Extraction(Extractable):
    """
    Encapsulates the results of an extractor.
    Note that Extractions are Extractable, so they can be used as inputs to other extractors.
    """

    def __init__(self,
                 value,
                 extractor_name: str,
                 confidence=1.0,
                 start_token=None,
                 end_token=None,
                 start_char=None,
                 end_char=None,
                 **options):
        Extractable.__init__(self)
        """

        Args:
            extracted_result (dict): the extracted result should be dict containing information like.
                value, extractor_name, confidence, start_token, end_token, start_char, end_char

        Returns:

        """
        self._addition_inf = dict()
        self._addition_inf["tag"] = options["tag"] if "tag" in options else None
        self._addition_inf["rule_id"] = options["rule_id"] if "rule_id" in options else None
        self._addition_inf["token_based_match_mapping"] = options["match_mapping"] if "match_mapping" in options else None
        self._addition_inf["date_object"] = options["date_object"] if "date_object" in options else None
        self._addition_inf["original_date"] = options["original_date"] if "original_date" in options else None
        self._extractor_name = extractor_name
        self._confidence = confidence
        self._provenance = {
            "start_token": start_token,
            "end_token": end_token,
            "start_char": start_char,
            "end_char": end_char,
            "extractor_name": extractor_name,
            "confidence": confidence
        }
        self._value = value

    def __str__(self):
        return str(self.__class__) + ", value: " + str(self._value)

    def __repr__(self):
        return str(self.__class__) + ", value: " + str(self._value)

    @property
    def value(self) -> Dict or str:
        """
        Returns: the value produced by an extractor
        """
        return self._value

    @property
    def confidence(self) -> float:
        """
        Returns: the confidence of this extraction
        """
        return self._confidence

    @property
    def name(self) -> str:
        """
        Returns: the name of this extraction
        """
        return self._extractor_name

    @property
    def tag(self) -> str:
        """

        Returns: the tag associated with this Extraction.

        """
        return self._addition_inf["tag"]

    @property
    def rule_id(self) -> str:
        """

        Returns: the rule_id associated with this Extraction.

        """
        return self._addition_inf["rule_id"]

    @property
    def token_based_match_mapping(self) -> Dict:
        """

        Returns: the token_based_match_mapping associated with this Extraction.

        """
        return self._addition_inf["token_based_match_mapping"]

    @property
    def original_date(self):
        """
        Returns: the original_date associated with this Extraction.
        """
        return self._addition_inf["original_date"]

    @property
    def date_object(self):
        """
        Returns: the original_date associated with this Extraction.
        """
        return self._addition_inf["date_object"]

    @property
    def provenance(self) -> Dict:
        """

        Returns: the tag associated with this Extraction.

        """
        return self._provenance



from typing import List
from etk.extraction import Extraction
from etk.extractor import Extractor, InputType

import copy
import spacy


class SentenceExtractor(Extractor):
    """
    **Description**
        Extract individual sentences using lightweight spaCy module.

    Example:
        ::

            sentence_extractor = SentenceExtractor(custom_nlp=nlp)
            sentence_extractor.extract(text=text)
    """

    def __init__(self, name: str = None, custom_nlp: type = None) -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="Text extractor",
                           name=name if name else "Sentence extractor")

        load_parser = False
        if custom_nlp:
            try:
                custom_pipeline = copy.deepcopy(custom_nlp)
                pipe_names = custom_pipeline.pipe_names
                for pipe in pipe_names:
                    if pipe != "parser":
                        custom_pipeline.remove_pipe(pipe)

                try:
                    assert "parser" in custom_pipeline.pipe_names
                    self._parser = custom_pipeline
                except AssertionError:
                    print("Note: custom_pipeline does not have a parser. \n"
                          "Loading parser from en_core_web_sm... ")
                    load_parser = True

            except AttributeError as e:
                print("Note: custom_pipeline does not have expected "
                      "attributes.")
                print(e)
                print("Loading parser from en_core_web_sm...")
                load_parser = True
        else:
            load_parser = True

        if load_parser:
            self._parser = spacy.load("en_core_web_sm",
                                     disable=["tagger", "ner"])

    def extract(self, text: str) -> List[Extraction]:
        """
        Splits text by sentences.

        Args:
            text (str): Input text to be extracted.

        Returns:
            List[Extraction]: the list of extraction or the empty list if there are no matches.
        """

        doc = self._parser(text)

        extractions = list()
        for sent in doc.sents:
            this_extraction = Extraction(value=sent.text,
                                         extractor_name=self.name,
                                         start_token=sent[0],
                                         end_token=sent[-1],
                                         start_char=sent.text[0],
                                         end_char=sent.text[-1])
            extractions.append(this_extraction)

        return extractions



import unittest
from etk.tokenizer import Tokenizer
import re


class TestTokenizer(unittest.TestCase):

    def test_tokenizer(self) -> None:
        text = "dsa@isi.edu 32.4 -32.1 (123)-345-6789, #1  \n \n   "
        reconstruct_text = re.sub(' +', ' ', text)
        t = Tokenizer()
        t.keep_multi_space = False
        tokens = t.tokenize(text)
        token_attrs = []
        for i in tokens:
            token_attrs.append({"orth": i.orth_, "offset": i.idx, "full_shape": i._.full_shape})
        expected = [
            {'orth': 'dsa', 'offset': 0, 'full_shape': 'xxx'},
            {'orth': '@', 'offset': 3, 'full_shape': '@'},
            {'orth': 'isi', 'offset': 4, 'full_shape': 'xxx'},
            {'orth': '.', 'offset': 7, 'full_shape': '.'},
            {'orth': 'edu', 'offset': 8, 'full_shape': 'xxx'},
            {'orth': '32.4', 'offset': 12, 'full_shape': 'dd.d'},
            {'orth': '-', 'offset': 17, 'full_shape': '-'},
            {'orth': '32.1', 'offset': 18, 'full_shape': 'dd.d'},
            {'orth': '(', 'offset': 23, 'full_shape': '('},
            {'orth': '123', 'offset': 24, 'full_shape': 'ddd'},
            {'orth': ')', 'offset': 27, 'full_shape': ')'},
            {'orth': '-', 'offset': 28, 'full_shape': '-'},
            {'orth': '345', 'offset': 29, 'full_shape': 'ddd'},
            {'orth': '-', 'offset': 32, 'full_shape': '-'},
            {'orth': '6789', 'offset': 33, 'full_shape': 'dddd'},
            {'orth': ',', 'offset': 37, 'full_shape': ','},
            {'orth': '#', 'offset': 39, 'full_shape': '#'},
            {'orth': '1', 'offset': 40, 'full_shape': 'd'},
            {'orth': '\n ', 'offset': 42, 'full_shape': '\n '},
            {'orth': '\n ', 'offset': 44, 'full_shape': '\n '}
        ]

        self.assertEqual(token_attrs, expected)
        self.assertEqual(t.reconstruct_text(tokens), reconstruct_text)

if __name__ == '__main__':
    unittest.main()


from typing import List
import re
from jsonpath_ng import jsonpath, parse
from etk.document import Document


class DocumentSelector(object):
    """
    Abstract class to encapsulate logic to enable ETK to selectively apply different extractors to different
    documents. Enables users of ETK to write custom logic for selecting documents for processing.
    """

    def select_document(self, document: Document) -> bool:
        """
        Args:
            document (Document): the document to be tested

        Returns: True, if the document should be selected for processing

        """


class DefaultDocumentSelector(DocumentSelector):

    def __init__(self):
        pass

    """
    A concrete implementation of DocumentSelector that supports commonly used methods for
    selecting documents.
    """

    """
        Args:
            document ():
            datasets (List[str]): test the "dataset" attribute in the doc contains any of the strings provided
            url_patterns (List[str]): test the "url" of the doc matches any of the regexes using regex.search
            website_patterns (List[str]): test the "website" of the doc contains any of the regexes using regex.search
            json_paths (List[str]): test existence of any of the given JSONPaths in a document
            json_paths_regex(List[str]): test that any of the values selected in 'json_paths' satisfy any of
              the regexex provided using regex.search

        Returns: True if the document satisfies all non None arguments.
        Each of the arguments can be a list, in which case satisfaction means
        that one of the elements in the list is satisfied, i.e., it is an AND of ORs

        For efficiency, the class caches compiled representations of all regexes and json_paths given
        that the same selectors will be used for consuming all documents in a stream.
        """

    # re1 = r'\d+\.\d*[L][-]\d*\s[A-Z]*[/]\d*'
    # re2 = '\d*[/]\d*[A-Z]*\d*\s[A-Z]*\d*[A-Z]*'
    # re3 = '[A-Z]*\d+[/]\d+[A-Z]\d+'
    # re4 = '\d+[/]\d+[A-Z]*\d+\s\d+[A-z]\s[A-Z]*'

    # sentences = [string1, string2, string3, string4]
    # generic_re = re.compile("(%s|%s|%s|%s)" % (re1, re2, re3, re4)).findall(sentence)

    def select_document(self,
                        document: Document,
                        datasets: List[str] = None,
                        url_patterns: List[str] = None,
                        website_patterns: List[str] = None,
                        json_paths: List[str] = None,
                        json_paths_regex: List[str] = None) -> bool:

        json_doc = document.cdr_document

        if (json_paths_regex is not None) and (json_paths is None):
            print("please specify both json_paths_regex and json_paths")
            # TODO: print out some error message here
            return False

        if datasets is not None:
            res = self.check_content(json_doc, "$.dataset", datasets, False)
            if not res:
                return False

        if url_patterns is not None:
            res = self.check_content(json_doc, "$.url", url_patterns, True)
            if not res:
                return False

        if website_patterns is not None:
            res = self.check_content(json_doc, "$.website", website_patterns, True)
            if not res:
                return False

        if json_paths is not None:
            rw_json_paths = []
            for json_path in json_paths:
                rw_json_paths.append(parse(json_path))

        if json_paths_regex is not None:
            rw_json_paths_regex = re.compile('|'.join(json_paths_regex))
            # TODO AMandeep: what is going at the line below
            res = self.check_json_path_codition(json_doc, rw_json_paths, rw_json_paths_regex)
            if not res:
                return False

        return True

    @staticmethod
    def check_content(json_doc: dict, json_path: str, patterns: List[str], enable_regexp: bool = True) -> bool:
        if enable_regexp:
            # compile json path
            patterns = re.compile('|'.join(patterns))
        else:
            patterns = '|'.join(patterns)

        rw_jsonpath = parse(json_path)

        # find all matches value
        values = [match.value for match in rw_jsonpath.find(json_doc)]
        # test if any matched value satisfies condition
        res = any(re.search(patterns, value) != None for value in values)

        return res

    @staticmethod
    def check_json_path_codition(json_doc: dict,
                                 rw_json_paths: List[jsonpath.Child],
                                 compiled_json_paths_regex: str) -> bool:
        for json_path_expr in rw_json_paths:
            values = [match.value for match in json_path_expr.find(json_doc)]
            res = any(re.search(compiled_json_paths_regex, value) is not None for value in values)
            if res:
                return True

        return False



from rdflib.graph import Graph, ConjunctiveGraph, Dataset, URIRef
from rdflib.plugins.sparql import prepareQuery
from rdflib.plugins.sparql.processor import SPARQLResult
import argparse
import warnings
import os
import hashlib


def read_by_chunk(f, chunk_size):
    i = 0
    # buff_prev + buff = slide window
    buff_prev = ''
    buff = ''

    for line in f:
        if chunk_size != 0:
            if i >= chunk_size:
                yield buff_prev + buff
                i = 0
                buff_prev = buff
                buff = ''
        buff += line + '\n'
        i += 1
    yield buff_prev + buff


def generate_index(triple):
    s, p, o = triple
    content = '{} {} {}'.format(str(s), str(p), str(o))
    return hashlib.md5(content.encode('utf-8')).hexdigest()


def add_arguments(parser):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    parser.description = 'Triple re-ontologization'
    parser.add_argument('-i', '--input-file', type=argparse.FileType('r'), dest='input_file')
    parser.add_argument('--input-type', action='store', dest='input_type', default='nt')
    parser.add_argument('-o', '--output-file', type=argparse.FileType('wb'), dest='output_file')
    parser.add_argument('--output-type', action='store', dest='output_type', default='nt')
    parser.add_argument('-q', '--query-file', type=argparse.FileType('r'), dest='query_file')
    parser.add_argument('-s', '--chunk-size', action='store', dest='chunk_size', default=0)
    parser.add_argument('-g', '--graph', type=argparse.FileType('r'), dest='graphs', nargs='+', default=[])


def run(args):
    """
    Args:
        args (argparse.Namespace)
    """

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        query = prepareQuery(args.query_file.read())
        ds = Dataset()

        res_indices_prev = set()  # de-duplication
        res_indices = set()

        # create sub graphs
        for f in args.graphs:
            g = Graph(identifier=os.path.basename(f.name))
            g.parse(data=f.read(), format='n3')
            ds.add_graph(g)

        # create and query data graph
        for data in read_by_chunk(args.input_file, int(args.chunk_size)):
            g = Graph(identifier='data')
            g.parse(data=data, format=args.input_type)
            ds.add_graph(g)
            res = ds.query(query)

            dedup_res_graph = Graph()
            if len(res) != 0:
                for r in res:
                    tid = generate_index(r)
                    res_indices.add(tid)
                    if tid in res_indices_prev:  # duplicated
                        continue
                    dedup_res_graph.add(r)

                if len(dedup_res_graph) > 0:
                    ret = dedup_res_graph.serialize(format=args.output_type, encoding='utf-8')
                    args.output_file.write(ret)

            ds.remove_graph(g)
            res_indices_prev = res_indices
            res_indices = set()



def tf_transfer(x):
    if isinstance(x, str):
        if x.lower() == "true":
            return True
        if x.lower() == "false":
            return False
        return True
    else:
        return x


import platform
import tempfile
from typing import List, Dict
import spacy, copy, json, os, jsonpath_ng.ext, importlib, logging, sys
from etk.tokenizer import Tokenizer
from etk.crf_tokenizer import CrfTokenizer
from etk.document import Document
from etk.etk_exceptions import InvalidJsonPathError
from etk.etk_module import ETKModule
from etk.etk_exceptions import ErrorPolicy, NotGetETKModuleError
from etk.utilities import Utility
import gzip

TEMP_DIR = '/tmp' if platform.system() == 'Darwin' else tempfile.gettempdir()


class ETK(object):
    def __init__(self, kg_schema=None, modules=None, extract_error_policy="process", logger=None,
                 logger_path=os.path.join(TEMP_DIR, 'etk.log'), ontology=None, generate_json_ld=False,
                 output_kg_only=False, use_spacy_tokenizer=False):

        self.generate_json_ld = generate_json_ld
        self.output_kg_only = output_kg_only

        if logger:
            self.logger = logger
        else:
            logging.basicConfig(
                level=logging.DEBUG,
                format='%(asctime)s %(name)-6s %(levelname)s %(message)s',
                datefmt='%m-%d %H:%M',
                filename=logger_path,
                filemode='w'
            )
            self.logger = logging.getLogger('ETK')

        self.parser = jsonpath_ng.ext.parse
        self.default_nlp = spacy.load('en_core_web_sm')
        if use_spacy_tokenizer:
            self.default_tokenizer = Tokenizer(copy.deepcopy(self.default_nlp))
        else:
            self.default_tokenizer = CrfTokenizer()
        self.parsed = dict()
        self.kg_schema = kg_schema
        self.ontology = ontology
        self.em_lst = list()
        if modules:
            if isinstance(modules, list):
                for module in modules:
                    if isinstance(module, str):
                        self.em_lst.extend(self.load_ems(modules))
                    elif issubclass(module, ETKModule):
                        self.em_lst.append(module(self))
            elif issubclass(modules, ETKModule):
                self.em_lst = [modules(self)]
            else:
                raise NotGetETKModuleError("Not getting extraction module")

        if extract_error_policy.lower() == "throw_extraction":
            self.error_policy = ErrorPolicy.THROW_EXTRACTION
        if extract_error_policy.lower() == "throw_document":
            self.error_policy = ErrorPolicy.THROW_DOCUMENT
        if extract_error_policy.lower() == "raise_error":
            self.error_policy = ErrorPolicy.RAISE
        else:
            self.error_policy = ErrorPolicy.PROCESS

    def create_document(self, doc: Dict, mime_type: str = None, url: str = "http://ex.com/123",
                        doc_id=None, type_=None) -> Document:
        """
        Factory method to wrap input JSON docs in an ETK Document object.

        Args:
            doc (object): a JSON object containing a document in CDR format.
            mime_type (str): if doc is a string, the mime_type tells what it is
            url (str): if the doc came from the web, specifies the URL for it
            doc_id
            type_

        Returns: wrapped Document

        """
        return Document(self, doc, mime_type, url, doc_id=doc_id).with_type(type_)

    def parse_json_path(self, jsonpath):

        """
        Parse a jsonpath

        Args:
            jsonpath: str

        Returns: a parsed json path

        """
        if jsonpath not in self.parsed:
            try:
                self.parsed[jsonpath] = self.parser(jsonpath)
            except Exception:
                self.log("Invalid Json Path: " + jsonpath, "error")
                raise InvalidJsonPathError("Invalid Json Path")

        return self.parsed[jsonpath]

    def process_and_frame(self, doc: Document):
        """
        Processes a document and if it has child docs, embeds them in the parent document. Only works for 1 level of
        nesting. Kind of hack, will implement properly later
        Args:
            doc: input document to be run etk modules on

        Returns:

        """
        nested_docs = self.process_ems(doc)
        parent_kg = doc.cdr_document.get('knowledge_graph', None)
        if parent_kg:
            if nested_docs and len(nested_docs) > 0:
                for nested_doc in nested_docs:
                    json_doc = nested_doc.cdr_document
                    doc_id = json_doc['doc_id']
                    if doc_id != doc.doc_id:
                        for field_name in list(parent_kg):
                            field_extractions = parent_kg[field_name]
                            if not isinstance(field_extractions, list):
                                field_extractions = [field_extractions]
                            for i in range(0, len(field_extractions)):
                                field_extraction = field_extractions[i]
                                if 'value' in field_extraction and field_extraction['value'] == doc_id:
                                    del field_extractions[i]
                                    field_extractions.append(
                                        {'value': json_doc, 'key': field_extraction['key'], 'is_nested': True})

    def process_ems(self, doc: Document) -> List[Document]:
        """
        Factory method to wrap input JSON docs in an ETK Document object.

        Args:
            doc (Document): process on this document

        Returns: a Document object and a KnowledgeGraph object

        """
        new_docs = list()

        for a_em in self.em_lst:
            if a_em.document_selector(doc):
                self.log(" processing with " + str(type(a_em)) + ". Process", "info", doc.doc_id, doc.url)
                fresh_docs = a_em.process_document(doc)
                # Allow ETKModules to return nothing in lieu of an empty list (people forget to return empty list)
                if fresh_docs:
                    new_docs.extend(fresh_docs)
            # try:
            #     if a_em.document_selector(doc):
            #         self.log(" processing with " + str(type(a_em)) + ". Process", "info", doc.doc_id, doc.url)
            #         new_docs.extend(a_em.process_document(doc))
            # except Exception as e:
            #     if self.error_policy == ErrorPolicy.THROW_EXTRACTION:
            #         self.log(str(e) + " processing with " + str(type(a_em)) + ". Continue", "error", doc.doc_id,
            #                  doc.url)
            #         continue
            #     if self.error_policy == ErrorPolicy.THROW_DOCUMENT:
            #         self.log(str(e) + " processing with " + str(type(a_em)) + ". Throw doc", "error", doc.doc_id,
            #                  doc.url)
            #         return list()
            #     if self.error_policy == ErrorPolicy.RAISE:
            #         self.log(str(e) + " processing with " + str(type(a_em)), "error", doc.doc_id, doc.url)
            #         raise e

        # Do house cleaning.
        doc.insert_kg_into_cdr()
        if not self.generate_json_ld:
            if "knowledge_graph" in doc.cdr_document:
                doc.cdr_document["knowledge_graph"].pop("@context", None)
        Utility.make_json_serializable(doc.cdr_document)

        if self.output_kg_only:
            doc = doc.kg.value
        elif not doc.doc_id:
            doc.doc_id = Utility.create_doc_id_from_json(doc.cdr_document)

        results = [doc]
        for new_doc in new_docs:
            results.extend(self.process_ems(new_doc))

        return results

    @staticmethod
    def load_glossary(file_path: str, read_json=False) -> List[str]:
        """
        A glossary is a text file, one entry per line.

        Args:
            file_path (str): path to a text file containing a glossary.
            read_json (bool): set True if the glossary is in json format
        Returns: List of the strings in the glossary.
        """
        if read_json:
            if file_path.endswith(".gz"):
                return json.load(gzip.open(file_path))
            return json.load(open(file_path))

        return open(file_path).read().splitlines()

    @staticmethod
    def load_spacy_rule(file_path: str) -> Dict:
        """
        A spacy rule file is a json file.

        Args:
            file_path (str): path to a text file containing a spacy rule sets.

        Returns: Dict as the representation of spacy rules
        """
        with open(file_path) as fp:
            return json.load(fp)

    def load_ems(self, modules_paths: List[str]):
        """
        Load all extraction modules from the path

        Args:
            modules_path: str

        Returns:

        """
        all_em_lst = []
        if modules_paths:
            for modules_path in modules_paths:
                em_lst = []
                try:
                    for file_name in os.listdir(modules_path):
                        if file_name.startswith("em_") and file_name.endswith(".py"):
                            sys.path.append(modules_path)  # append module dir path
                            this_module = importlib.import_module(file_name[:-3])
                            for em in self.classes_in_module(this_module):
                                em_lst.append(em(self))
                except:
                    self.log("Error when loading etk modules from " + modules_path, "error")
                    raise NotGetETKModuleError("Wrong file path for ETK modules")
                all_em_lst += em_lst

        try:
            all_em_lst = self.topological_sort(all_em_lst)
        except Exception:
            self.log("Topological sort for ETK modules fails", "error")
            raise NotGetETKModuleError("Topological sort for ETK modules fails")

        # if not all_em_lst:
        #     self.log("No ETK module in " + str(modules_paths), "error")
        #     raise NotGetETKModuleError("No ETK module in dir, module file should start with em_, end with .py")
        return all_em_lst

    @staticmethod
    def topological_sort(etk_module_list: List[ETKModule]) -> List[ETKModule]:
        """
        Return topological order of ems

        Args:
            lst: List[ExtractionModule]

        Returns: List[ExtractionModule]

        """
        "TODO"
        return etk_module_list

    @staticmethod
    def classes_in_module(module) -> List:
        """
        Return all classes with super class ExtractionModule

        Args:
            module:

        Returns: List of classes

        """
        md = module.__dict__
        return [
            md[c] for c in md if (
                    isinstance(md[c], type) and
                    issubclass(md[c], ETKModule
                               ) and
                    md[c].__module__ == module.__name__)
        ]

    def log(self, message, level, doc_id=None, url=None):
        message = message + " doc_id: {}".format(doc_id) + " url: {}".format(url)

        if level == "error":
            self.logger.error(message)
        elif level == "warning":
            self.logger.warning(message)
        elif level == "info":
            self.logger.info(message)
        elif level == "debug":
            self.logger.debug(message)
        elif level == "critical":
            self.logger.critical(message)
        elif level == "exception":
            self.logger.exception(message)



#!/usr/bin/env python
import logging
import re
import sys

import cssselect
from collections import defaultdict
from collections import deque
from lxml.etree import tostring
from lxml.etree import tounicode
from lxml.html import document_fromstring
from lxml.html import fragment_fromstring

from .cleaners import clean_attributes
from .cleaners import html_cleaner
from .htmls import build_doc
from .htmls import get_body
from .htmls import get_title
from .htmls import shorten_title


logging.basicConfig(level=logging.INFO)
log = logging.getLogger()

REGEXES = {
    'unlikelyCandidatesRe': re.compile('combx|comment|community|disqus|extra|foot|header|menu|remark|rss|shoutbox|sidebar|sponsor|ad-break|agegate|pagination|pager|popup|tweet|twitter', re.I),
    'okMaybeItsACandidateRe': re.compile('and|article|body|column|main|shadow', re.I),
    'positiveRe': re.compile('ad-post|category|article|body|content|entry|hentry|main|page|pagination|post|text|blog|story', re.I),
    'negativeRe': re.compile('combx|comment|com-|contact|foot|footer|footnote|masthead|media|meta|outbrain|promo|related|scroll|shoutbox|sidebar|sponsor|shopping|tags|tool|widget', re.I),
    'divToPElementsRe': re.compile('<(a|blockquote|dl|div|img|ol|p|pre|table|ul)', re.I),
    'socialNetworksRe': re.compile('facebook|twitter|disqus|tumblr|linkedin|flickr'),
    #'replaceBrsRe': re.compile('(<br[^>]*>[ \n\r\t]*){2,}',re.I),
    #'replaceFontsRe': re.compile('<(\/?)font[^>]*>',re.I),
    #'trimRe': re.compile('^\s+|\s+$/'),
    #'normalizeRe': re.compile('\s{2,}/'),
    #'killBreaksRe': re.compile('(<br\s*\/?>(\s|&nbsp;?)*){1,}/'),
    'videoRe': re.compile('http:\/\/(www\.)?(youtube|vimeo)\.com', re.I),
    #skipFootnoteLink:      /^\s*(\[?[a-z0-9]{1,2}\]?|^|edit|citation needed)\s*$/i,
}


class Unparseable(ValueError):
    pass

def describe(node, depth=1):
    if not hasattr(node, 'tag'):
        return "[%s]" % type(node)
    name = node.tag
    if node.get('id', ''):
        name += '#' + node.get('id')
    if node.get('class', ''):
        name += '.' + node.get('class').replace(' ', '.')
    if name[:4] in ['div#', 'div.']:
        name = name[3:]
    if depth and node.getparent() is not None:
        return name + ' - ' + describe(node.getparent(), depth - 1)
    return name


def to_int(x):
    if not x:
        return None
    x = x.strip()
    if x.endswith('px'):
        return int(x[:-2])
    if x.endswith('em'):
        return int(x[:-2]) * 12
    return int(x)


def clean(text):
    text = re.sub('\s*\n\s*', '\n', text)
    text = re.sub('[ \t]{2,}', ' ', text)
    return text.strip()


def text_length(i):
    return len(clean(i.text_content() or ""))

regexp_type = type(re.compile('hello, world'))

def compile_pattern(elements):
    if not elements:
        return None
    if isinstance(elements, regexp_type):
        return elements
    if isinstance(elements, str):
        elements = elements.split(',')
    return re.compile(u'|'.join([re.escape(x.lower()) for x in elements]), re.U)

class Document:
    """Class to build a etree document out of html."""
    TEXT_LENGTH_THRESHOLD = 5
    RETRY_LENGTH = 25

    def __init__(self, input, positive_keywords=None, negative_keywords=None, recallPriority=False, **options):
        """Generate the document

        :param input: string of the html content.

        kwargs:
            - attributes:
            - debug: output debug messages
            - min_text_length:
            - retry_length:
            - url: will allow adjusting links to be absolute
            - positive_keywords: the list of positive search patterns in classes and ids, for example: ["news-item", "block"]
            - negative_keywords: the list of negative search patterns in classes and ids, for example: ["mysidebar", "related", "ads"]
            Also positive_keywords and negative_keywords could be a regexp.
        """
        self.input = input
        self.recallPriority = recallPriority
        self.options = options
        self.html = None
        self.encoding = None
        self.positive_keywords = compile_pattern(positive_keywords)
        self.negative_keywords = compile_pattern(negative_keywords)

    def _html(self, force=False):
        if force or self.html is None:
            self.html = self._parse(self.input)
        return self.html

    def _parse(self, input):
        doc, self.encoding = build_doc(input)
        doc = html_cleaner.clean_html(doc)
        base_href = self.options.get('url', None)
        if base_href:
            doc.make_links_absolute(base_href, resolve_base_href=True)
        else:
            doc.resolve_base_href()
        return doc

    def content(self):
        return get_body(self._html(True))

    def title(self):
        return get_title(self._html(True))

    def short_title(self):
        return shorten_title(self._html(True))

    def get_clean_html(self):
         return clean_attributes(tounicode(self.html))

    def summary(self, html_partial=False):
        """Generate the summary of the html docuemnt

        :param html_partial: return only the div of the document, don't wrap
        in html and body tags.

        """
        try:
            ruthless = True
            
            #Added recall priority flag
            recallPriority = self.recallPriority
            if recallPriority:
                ruthless = False
                self.TEXT_LENGTH_THRESHOLD = 2
                self.RETRY_LENGTH = 25

            while True:
                self._html(True)

                for i in self.tags(self.html, 'script', 'style'):
                    i.drop_tree()
                for i in self.tags(self.html, 'body'):
                    i.set('id', 'readabilityBody')

                if ruthless:
                    self.remove_unlikely_candidates()

                self.transform_misused_divs_into_paragraphs()
                candidates = self.score_paragraphs()

                best_candidates = self.select_best_candidates(candidates)


                if best_candidates and not recallPriority:
                    article = self.get_article_from_candidates(candidates,best_candidates,html_partial)
                else:
                    if ruthless and not recallPriority:
                        log.debug("ruthless removal did not work. ")
                        ruthless = False
                        self.debug(
                            ("ended up stripping too much - "
                             "going for a safer _parse"))
                        # try again
                        continue
                    else:

                        log.debug(
                            ("Ruthless and lenient parsing did not work. "
                             "Returning raw html"))
                        article = self.html.find('body')
                        if article is None:
                            article = self.html

                cleaned_article = self.sanitize(article, candidates)
                # print(cleaned_article)

                article_length = len(cleaned_article or '')
                retry_length = self.options.get(
                    'retry_length',
                    self.RETRY_LENGTH)

                of_acceptable_length = article_length >= retry_length
                if ruthless and not of_acceptable_length:
                    ruthless = False
                    continue
                else:
                    return cleaned_article
        except Exception as e:
            print("error: %s", e)
            log.exception('error getting summary: ')
            raise Exception(Unparseable(str(e)), None, sys.exc_info()[2])

    def get_article_from_candidates(self,candidates,best_candidates,html_partial=False):
        nodes = []
        if html_partial:
            output = fragment_fromstring('<div/>')
        else:
            output = document_fromstring('<div/>')

        for best_candidate in best_candidates:
            article, nodes = self.get_article(candidates,best_candidate,nodes,html_partial)
            output.append(article)

        return output

    def get_article(self, candidates, best_candidate,nodes, html_partial=False):
        #print nodes
        # Now that we have the top candidate, look through its siblings for
        # content that might also be related.
        # Things like preambles, content split by ads that we removed, etc.
        sibling_score_threshold = max([
            10,
            best_candidate['content_score'] * 0.2])
        # create a new html document with a html->body->div
        if html_partial:
            output = fragment_fromstring('<div/>')
        else:
            output = document_fromstring('<div/>')
        best_elem = best_candidate['elem']
        parent = best_elem.getparent()
        
        siblings = parent.getchildren() if parent is not None else [best_elem]
        for sibling in siblings:
            # in lxml there no concept of simple text
            # if isinstance(sibling, NavigableString): continue
            append = False
            if sibling is best_elem:
                append = True
            sibling_key = sibling  # HashableElement(sibling)
            if sibling_key in candidates and \
                candidates[sibling_key]['content_score'] >= sibling_score_threshold:
                append = True

            if sibling.tag == "p":
                link_density = self.get_link_density(sibling)
                node_content = sibling.text or ""
                node_length = len(node_content)

                if node_length > 20 and link_density < 0.25:
                    append = True
                elif node_length <= 20 \
                    and link_density == 0 \
                    and re.search('\.( |$)', node_content):
                    append = True

            if append:
                if not sibling in nodes:
                    nodes += self.get_all_child_nodes(sibling)
                    if html_partial:
                        output.append(sibling)
                    else:
                        output.getchildren()[0].getchildren()[0].append(sibling)

        return output,nodes

    def get_all_child_nodes(self,candidate):
        #Gets all child nodes
        queue,nodes = deque(),[]
        for sibling in candidate.getchildren():
            queue.append(sibling)

        while queue:
            current = queue.popleft()
            for sib in current.getchildren():
                queue.append(sib)

            nodes.append(current)
        return nodes 

    def select_best_candidates(self,candidates):
        #Gets all candidate
        sorted_candidates = sorted(candidates.values(), key=lambda x: x['content_score'], reverse=True)
        if len(sorted_candidates) == 0:
            return None
        
        output = []
        for candidate in sorted_candidates:
            if not candidate['content_score'] >0:
                break
            output.append(candidate)
        return output

    def select_best_candidate(self, candidates):
        sorted_candidates = sorted(candidates.values(), key=lambda x: x['content_score'], reverse=True)
        for candidate in sorted_candidates[:5]:
            elem = candidate['elem']
            self.debug("Top 5 : %6.3f %s" % (
                candidate['content_score'],
                describe(elem)))

        if len(sorted_candidates) == 0:
            return None

        best_candidate = sorted_candidates[0]
        return best_candidate

    def get_link_density(self, elem):
        link_length = 0
        for i in elem.findall(".//a"):
            link_length += text_length(i)

        total_length = text_length(elem)
        return float(link_length) / max(total_length, 1)

    def score_paragraphs(self, ):
        MIN_LEN = self.options.get(
            'min_text_length',
            self.TEXT_LENGTH_THRESHOLD)
        candidates = {}
        ordered = []
        for elem in self.tags(self._html(), "p", "pre", "td", "ul"):

            parent_node = elem.getparent()
            if parent_node is None:
                continue
            grand_parent_node = parent_node.getparent()

            inner_text = clean(elem.text_content() or "")
            inner_text_len = len(inner_text)

            # If this paragraph is less than 25 characters
            # don't even count it.
            if inner_text_len < MIN_LEN:
                continue

            if parent_node not in candidates:
                candidates[parent_node] = self.score_node(parent_node)
                ordered.append(parent_node)

            if grand_parent_node is not None and grand_parent_node not in candidates:
                candidates[grand_parent_node] = self.score_node(
                    grand_parent_node)
                ordered.append(grand_parent_node)

            content_score = 1
            content_score += len(inner_text.split(','))
            content_score += min((inner_text_len / 100), 3)

            #WTF? candidates[elem]['content_score'] += content_score
            candidates[parent_node]['content_score'] += content_score
            if grand_parent_node is not None:
                candidates[grand_parent_node]['content_score'] += content_score / 2.0

        # Scale the final candidates score based on link density. Good content
        # should have a relatively small link density (5% or less) and be
        # mostly unaffected by this operation.
        for elem in ordered:
            candidate = candidates[elem]
            ld = self.get_link_density(elem)
            score = candidate['content_score']
            self.debug("Candid: %6.3f %s link density %.3f -> %6.3f" % (
                score,
                describe(elem),
                ld,
                score * (1 - ld)))
            candidate['content_score'] *= (1 - ld)

        return candidates

    def class_weight(self, e):
        weight = 0
        for feature in [e.get('class', None), e.get('id', None)]:
            if feature:
                if REGEXES['negativeRe'].search(feature):
                    weight -= 25

                if REGEXES['positiveRe'].search(feature):
                    weight += 25

                if self.positive_keywords and self.positive_keywords.search(feature):
                    weight += 25

                if self.negative_keywords and self.negative_keywords.search(feature):
                    weight -= 25

        if self.positive_keywords and self.positive_keywords.match('tag-'+e.tag):
            weight += 25

        if self.negative_keywords and self.negative_keywords.match('tag-'+e.tag):
            weight -= 25

        return weight

    def score_node(self, elem):
        content_score = self.class_weight(elem)
        name = elem.tag.lower()
        if name == "div":
            content_score += 5
        elif name in ["pre", "td", "blockquote"]:
            content_score += 3
        elif name in ["address", "ol", "ul", "dl", "dd", "dt", "li", "form"]:
            content_score -= 3
        elif name in ["h1", "h2", "h3", "h4", "h5", "h6", "th"]:
            content_score -= 5
        return {
            'content_score': content_score,
            'elem': elem
        }

    def debug(self, *a):
        if self.options.get('debug', False):
            log.debug(*a)

    def remove_unlikely_candidates(self):
        for elem in self.html.iter():
            s = "%s %s" % (elem.get('class', ''), elem.get('id', ''))
            if len(s) < 2:
                continue

            if REGEXES['unlikelyCandidatesRe'].search(s) and (not REGEXES['okMaybeItsACandidateRe'].search(s)) and elem.tag not in ['html', 'body']:
                self.debug("Removing unlikely candidate - %s" % describe(elem))
                elem.drop_tree()

    def transform_misused_divs_into_paragraphs(self):
        for elem in self.tags(self.html, 'div'):
            # transform <div>s that do not contain other block elements into
            # <p>s
            #FIXME: The current implementation ignores all descendants that
            # are not direct children of elem
            # This results in incorrect results in case there is an <img>
            # buried within an <a> for example
            if not REGEXES['divToPElementsRe'].search(
                    ''.join(map(lambda x: x.decode("utf-8"), map(tostring, list(elem))))):
                #self.debug("Altering %s to p" % (describe(elem)))
                elem.tag = "p"
                #print "Fixed element "+describe(elem)

        for elem in self.tags(self.html, 'div'):
            if elem.text and elem.text.strip():
                p = fragment_fromstring('<p/>')
                p.text = elem.text
                elem.text = None
                elem.insert(0, p)
                #print "Appended "+tounicode(p)+" to "+describe(elem)

            for pos, child in reversed(list(enumerate(elem))):
                if child.tail and child.tail.strip():
                    p = fragment_fromstring('<p/>')
                    p.text = child.tail
                    child.tail = None
                    elem.insert(pos + 1, p)
                    #print "Inserted "+tounicode(p)+" to "+describe(elem)
                if child.tag == 'br':
                    #print 'Dropped <br> at '+describe(elem)
                    child.drop_tree()

    def tags(self, node, *tag_names):
        for tag_name in tag_names:
            for e in node.findall('.//%s' % tag_name):
                yield e

    def reverse_tags(self, node, *tag_names):
        for tag_name in tag_names:
            for e in reversed(node.findall('.//%s' % tag_name)):
                yield e

    def sanitize(self, node, candidates):
        MIN_LEN = self.options.get('min_text_length',
            self.TEXT_LENGTH_THRESHOLD)
        for header in self.tags(node, "h1", "h2", "h3", "h4", "h5", "h6"):
            if self.class_weight(header) < 0 or self.get_link_density(header) > 0.33:
                header.drop_tree()

        for elem in self.tags(node, "form", "textarea"):
            elem.drop_tree()

        for elem in self.tags(node, "iframe"):
            if "src" in elem.attrib and REGEXES["videoRe"].search(elem.attrib["src"]):
                elem.text = "VIDEO" # ADD content to iframe text node to force <iframe></iframe> proper output
            else:
                elem.drop_tree()

        allowed = {}
        # Conditionally clean <table>s, <ul>s, and <div>s
        for el in self.reverse_tags(node, "table", "ul", "div"):
            if el in allowed:
                continue
            weight = self.class_weight(el)
            if el in candidates:
                content_score = candidates[el]['content_score']
                #print '!',el, '-> %6.3f' % content_score
            else:
                content_score = 0
            tag = el.tag

            if weight + content_score < 0:
                self.debug("Cleaned %s with score %6.3f and weight %-3s" %
                    (describe(el), content_score, weight, ))
                el.drop_tree()
            elif el.text_content().count(",") < 10:
                counts = {}
                for kind in ['p','td', 'img', 'li', 'a', 'embed', 'input']:
                    counts[kind] = len(el.findall('.//%s' % kind))
                counts["li"] -= 100
                counts["input"] -= len(el.findall('.//input[@type="hidden"]'))

                # Count the text length excluding any surrounding whitespace
                content_length = text_length(el)
                link_density = self.get_link_density(el)
                parent_node = el.getparent()
                if parent_node is not None:
                    if parent_node in candidates:
                        content_score = candidates[parent_node]['content_score']
                    else:
                        content_score = 0
                to_remove = False
                reason = ""

                if el.tag == 'div' and counts["img"] >= 1 and content_length == 0:
                    reason = "Just images (%s)" % counts["img"]
                    #print reason,el
                    to_remove = True
                # if el.tag == "ul":
                #     print el.text_content()
                 #    continue
                if counts["td"] and counts["img"] > 1+counts["td"]*1.3:
                    reason = "too many images (%s)" % counts["img"]
                    to_remove = True
                elif counts["li"] > counts["p"] and tag != "ul" and tag != "ol":
                    reason = "more <li>s than <p>s"
                    to_remove = True
                elif counts["input"] > (counts["p"] / 3):
                    reason = "less than 3x <p>s than <input>s"
                    to_remove = True
                elif content_length < (MIN_LEN) and (counts["img"] == 0 or counts["img"] > 2):
                    reason = "too short content length %s without a single image" % content_length
                    to_remove = True
                elif weight < 25 and link_density > 0.2:
                        reason = "too many links %.3f for its weight %s" % (
                            link_density, weight)
                        to_remove = True
                elif weight >= 25 and link_density > 0.5:
                    reason = "too many links %.3f for its weight %s" % (
                        link_density, weight)
                    to_remove = True
                elif (counts["embed"] == 1 and content_length < 75) or counts["embed"] > 1:
                    reason = "<embed>s with too short content length, or too many <embed>s"
                    to_remove = True

                #Removes all social network link matches and removes the links from output
                if el.findall('.//a'):
                    for ele in el.cssselect("a"):
                        link = ele.get('href') 
                        try:
                            if link is not None:
                                if REGEXES['socialNetworksRe'].search(link):
                                    ele.drop_tree()
                        except Exception as e:
                            print("Skip:", link, "cant be parsed")
                        

                if el.tag == 'div' and counts['img'] >= 1 and to_remove: 
                    imgs = el.findall('.//img')
                    valid_img = False
                    self.debug(tounicode(el))
                    for img in imgs:
                       height = img.get('height')
                       text_len = img.get('text_length')
                       self.debug ("height %s text_length %s" %(repr(height), repr(text_len)))
                       if type(to_int(height)) == int and to_int(text_len) == int:
                           if to_int(height) >= 100 or to_int(text_len) >= 25:
                               valid_img = True
                               self.debug("valid image" + tounicode(img))
                               break
                    if valid_img:
                       to_remove = False
                       #print ("Allowing %s" %el.text_content())
                       self.debug("Allowing %s" %el.text_content())
                       for desnode in self.tags(el, "table", "ul", "div"):
                           allowed[desnode] = True

                    #find x non empty preceding and succeeding siblings
                    i, j = 0, 0
                    x = 1
                    siblings = []
                    for sib in el.itersiblings():
                        #self.debug(sib.text_content())
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            i =+ 1
                            siblings.append(sib_content_length)
                            if i == x:
                                break
                    for sib in el.itersiblings(preceding=True):
                        #self.debug(sib.text_content())
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            j =+ 1
                            siblings.append(sib_content_length)
                            if j == x:
                                break
                    #self.debug(str(siblings))
                    if siblings and sum(siblings) > 1000:
                        to_remove = False
                        self.debug("Allowing %s" % describe(el))
                        for desnode in self.tags(el, "table", "ul", "div"):
                            allowed[desnode] = True

                if to_remove:
                    self.debug("Cleaned %6.3f %s with weight %s cause it has %s." %
                        (content_score, describe(el), weight, reason))

                    el.drop_tree()

        for el in ([node] + [n for n in node.iter()]):
            if not self.options.get('attributes', None):
                #el.attrib = {} #FIXME:Checkout the effects of disabling this
                pass

        self.html = node
        return self.get_clean_html()


class HashableElement():
    def __init__(self, node):
        self.node = node
        self._path = None

    def _get_path(self):
        if self._path is None:
            reverse_path = []
            node = self.node
            while node is not None:
                node_id = (node.tag, tuple(node.attrib.items()), node.text)
                reverse_path.append(node_id)
                node = node.getparent()
            self._path = tuple(reverse_path)
        return self._path
    path = property(_get_path)

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        return self.path == other.path

    def __getattr__(self, tag):
        return getattr(self.node, tag)


def main():
    from optparse import OptionParser
    parser = OptionParser(usage="%prog: [options] [file]")
    parser.add_option('-v', '--verbose', action='store_true')
    parser.add_option('-u', '--url', default=None, help="use URL instead of a local file")
    parser.add_option('-p', '--positive-keywords', default=None, help="positive keywords (separated with comma)", action='store')
    parser.add_option('-n', '--negative-keywords', default=None, help="negative keywords (separated with comma)", action='store')
    (options, args) = parser.parse_args()

    if not (len(args) == 1 or options.url):
        parser.print_help()
        sys.exit(1)

    file = None
    if options.url:
        import urllib.request
        file = urllib.request.urlopen(options.url)
    else:
        file = open(args[0], 'rt')
    enc = sys.__stdout__.encoding or 'utf-8' # XXX: this hack could not always work, better to set PYTHONIOENCODING
    try:
        print(Document(file.read(),
            debug=options.verbose,
            url=options.url,
            positive_keywords = options.positive_keywords,
            negative_keywords = options.negative_keywords,
        ).summary().encode(enc, 'replace'))
    finally:
        file.close()

if __name__ == '__main__':
    main()



from typing import List, Dict, Tuple

import copy
import re
import pyexcel

from etk.etk import ETK
from etk.extractor import Extractor, InputType
from etk.extraction import Extraction


class ExcelExtractor(Extractor):
    """
    **Description**
        This class is a wrapper of pyexcel. By defining 'variables' argument,
        this extractor is able to extract multiple information and store them
        into different field at once.

    Examples:
        ::

            excel_extractor = ExcelExtractor()
            excel_extractor.extract(file_name:path_to_excel,
                                    sheet_name:sheet_name,
                                    region:['A,1', 'Z,10'],
                                    variables={
                                        # get current col & row for field1
                                        'field1': '$col,$row',
                                        # get fixed cell of $A,$5 for field2
                                        'field2': '$A,$5'
                                        })

    """

    _re_row_identifier = re.compile(r'(\$[0-9]+)')
    _re_col_identifier = re.compile(r'(\$[A-Za-z]+)')

    def __init__(self, etk: ETK = None, extractor_name: str = 'excel extractor') -> None:
        Extractor.__init__(self,
                           input_type=InputType.TEXT,
                           category="data extractor",
                           name=extractor_name)
        self._etk = etk

    def extract(self, file_name: str, sheet_name: str, region: List, variables: Dict) -> List[Extraction]:
        """
        Args:
            file_name (str): file name
            sheet_name (str): sheet name
            region (List[]): from upper left cell to bottom right cell, e.g., ['A,1', 'Z,10']
            variables (Dict): key is variable name, value can be:
                              1. a single expression 2. comma separated expression, will be treated as location
                              $row, $col are built-in variables can be used in expression
                              constant row and column value can be noted as $NAME (e.g., $1, $10, $A, $GG)
        
        Returns:
            List[Extraction] : A list of extracted variables dictionary
            
        """
        extractions = []

        book = pyexcel.get_book(file_name=file_name)
        sheet = book[sheet_name]
        region = [ExcelExtractor._excel_coord_to_location(coord) for coord in region]
        r = region[0][0]
        # per row
        for row in sheet.region(region[0], region[1]):
            c = region[0][1]
            # per col
            for col in row:
                var = copy.deepcopy(variables)
                # per variable
                for k, v in var.items():
                    parsed_v = ExcelExtractor._parse_variable(v, r, c)
                    if len(parsed_v) == 1:  # normal variable
                        var[k] = parsed_v[0]
                    else:  # location
                        rr, cc = parsed_v
                        var[k] = sheet[rr, cc]
                    extractions.append(var)

                c += 1
            r += 1

        return extractions

    @staticmethod
    def _col_name_to_num(name: str) -> int:
        name = name.upper()
        pow = 1
        col_num = 0
        for letter in name[::-1]:
            col_num += (int(letter, 36) - 9) * pow
            pow *= 26
        return col_num - 1

    @staticmethod
    def _row_name_to_num(name: str) -> int:
        try:
            num = int(name) - 1
            if num >= 0:
                return num
            raise ValueError('Invalid row name')
        except:
            raise ValueError('Invalid row name')

    @staticmethod
    def _excel_coord_to_location(s: str) -> Tuple:
        ss = s.split(',')
        return ExcelExtractor._row_name_to_num(ss[1]), ExcelExtractor._col_name_to_num(ss[0])

    @staticmethod
    def _parse_variable(s: str, curr_row: int, curr_col: int) -> Tuple:
        '''
        $A,$2 <- constant col and row
        $row,$2 <- current col, row 2
        $A+1,$2 <- col A + 1 = 2, row 2
        $row+1,$2 <- current col + 1, row 2
        $A,$2-1 <-- col A, row 2 - 1 = 1
        '''

        def parse_expression(ss, curr_row, curr_col):
            ss = ss.replace('$row', str(curr_row))
            ss = ss.replace('$col', str(curr_col))
            ss = ExcelExtractor._re_row_identifier.sub(
                lambda x: str(ExcelExtractor._row_name_to_num(x.group()[1:])) if len(x.group()) > 0 else '', ss)
            ss = ExcelExtractor._re_col_identifier.sub(
                lambda x: str(ExcelExtractor._col_name_to_num(x.group()[1:])) if len(x.group()) > 0 else '', ss)
            return eval(ss)

        ss = s.split(',')
        if len(ss) == 1:
            return parse_expression(ss[0], curr_row, curr_col),
        elif len(ss) == 2:
            rr, cc = (ss[1], ss[0])
            return parse_expression(rr, curr_row, curr_col), parse_expression(cc, curr_row, curr_col)
        else:
            raise ValueError('Invalid variable')



import sys
import os
import re
import argparse
import pyexcel
import numpy
from datetime import date, datetime
import pyexcel
import logging
import demjson
import json
import copy
import etk.timeseries.location_range as lr
import etk.timeseries.location_parser as lp
import etk.timeseries.time_series_region as tsr

logging.basicConfig(format='%(name)s - %(levelname)s - %(message)s', level=logging.WARN)

class SpreadsheetAnnotation(object):
    def __init__(self, annotation, fn):
        self.locparser = lp.LocationParser()

        self.properties = annotation['Properties']
        self.sheet_indices = self.locparser.parse_range(annotation['Properties']['sheet_indices'])

        self.metadata = self.parse_md(annotation['GlobalMetadata'])
        self.provenance = dict(filename=fn)

        self.timeseries_regions = []
        for rgn in annotation['TimeSeriesRegions']:
            self.timeseries_regions.append(self.parse_tsr(rgn))

    def parse_md(self, md_json):
        md_dict = {}
        for mdspec in md_json:
            mdname = mdspec['name']
            md_dict[mdname] = {}
            md_dict[mdname]['source'] = mdspec['source']
            if mdspec['source'] == 'cell':
                (md_dict[mdname]['row'], md_dict[mdname]['col']) = self.locparser.parse_coords(mdspec['loc'])
            if mdspec['source'] == 'const':
                md_dict[mdname]['val'] = mdspec['val']
        return md_dict

    def parse_tsr(self, tsr_json):
        orientation = tsr_json['orientation']
        series_range = None
        if orientation == 'row':
            series_range = self.locparser.parse_range(tsr_json['rows'])
        else:
            series_range = self.locparser.parse_range(tsr_json['cols'])

        data_range = self.locparser.parse_range(tsr_json['locs'])

        time_coords = {}
        time_coords['locs'] = self.locparser.parse_range(tsr_json['times']['locs'])
        time_coords['granularity'] = tsr_json['times']['granularity'] if 'granularity' in tsr_json['times'] else 'unknown'
        if 'mode' in tsr_json['times']:
            time_coords['mode'] = tsr_json['times']['mode']
        else:
            time_coords['mode'] = None

        time_coords['post_process'] = tsr_json['times'].get('post_process')

        mdspec = self.parse_tsr_metadata(tsr_json['metadata'], orientation)

        return tsr.TimeSeriesRegion(orientation=orientation, series_range=series_range, data_range=data_range,
                                metadata_spec=mdspec, time_coordinates=time_coords, global_metadata=self.metadata,
                                provenance = self.provenance)

    def parse_tsr_metadata(self, md_json, orientation):
        md_dict = {}
        reverse_orientation = {'row': 'col', 'col': 'row'}
        for md_sec in md_json:
            name = md_sec['name']
            md_dict[name] = {}

            if 'source' in md_sec:
                md_dict[name]['source'] = md_sec['source']
            else:
                md_dict[name]['source'] = reverse_orientation[orientation]

            loc = None
            if 'loc' in md_sec:
                loc = md_sec['loc']

            if md_dict[name]['source'] == 'cell':
                md_dict[name]['loc'] = self.locparser.parse_coords(loc)

            elif md_dict[name]['source'] == 'row':
                md_dict[name]['loc'] = self.locparser.parse_range(loc)

            elif md_dict[name]['source'] == 'col':
                md_dict[name]['loc'] = self.locparser.parse_range(loc)

            elif md_dict[name]['source'] == 'const':
                md_dict[name]['val'] = md_sec['val']

            if 'mode' in md_sec:
                md_dict[name]['mode'] = md_sec['mode']
            else:
                md_dict[name]['mode'] = 'normal'

        return md_dict


class ExtractSpreadsheet(object):

    def __init__(self, spreadsheet_fn, annotations_fn):
        self.normalized_source_file = os.path.basename(spreadsheet_fn)
        self.book = pyexcel.get_book(file_name=spreadsheet_fn, auto_detect_datetime=False)
        self.annotations = self.load_annotations(annotations_fn)

    def process(self):
        timeseries = []
        for annotation in self.annotations:
            ssa = SpreadsheetAnnotation(annotation, self.normalized_source_file)
            parsed = []
            for anidx in ssa.sheet_indices:
                logging.debug("Parsing sheet %d", anidx)
                sheet = self.book.sheet_by_index(anidx)
                data = sheet.to_array()
                for rgn in ssa.timeseries_regions:
                    rgn.provenance['sheet']=anidx
                    for parsed_tsr in rgn.parse(data, sheet.name):
                        yield parsed_tsr
#                        parsed.append(parsed_tsr)
#                logging.debug("%s",parsed)
#            timeseries.append(parsed)
#        return timeseries
    def load_annotations(self,annotations_fn):
        anfile = open(annotations_fn)
        annotations_decoded = demjson.decode(anfile.read(), return_errors=True)
        for msg in annotations_decoded[1]:
            if msg.severity == "error":
                logging.error(msg.pretty_description())
        anfile.close()
        return annotations_decoded[0]


def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("annotation", help='Annotation of time series in custom JSON format')
    ap.add_argument("spreadsheet", help='Excel spreadsheet file')
    ap.add_argument("outfile", help='file to write results')
    args = ap.parse_args()
    es = ExtractSpreadsheet(args.spreadsheet, args.annotation)
    with open(args.outfile, 'w') as outfile:
            for timeseries in es.process():
                json.dump(timeseries, outfile)
                outfile.write("\n");

if __name__ == "__main__":
    main()



import sys
from datetime import date, datetime
import logging
import json
import copy
import decimal
import etk.timeseries.location_range as lr
import etk.timeseries.location_parser as lp
import hashlib
import re

logging.basicConfig(format='%(name)s - %(levelname)s - %(message)s', level=logging.WARN)


class DecimalJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalJSONEncoder, self).default(o)


class TimeSeriesRegion(object):
    def __init__(self, orientation='row', series_range=None, data_range=None, metadata_spec=None,
                 time_coordinates=None, global_metadata=None, granularity=None, provenance=None):
        self.orientation = orientation
        self.series_range = series_range
        self.data_range = data_range
        self.granularity = granularity
        self.metadata_spec = metadata_spec
        self.time_coordinates = time_coordinates
        self.global_metadata = global_metadata
        self.provenance = provenance
        self.time_series = []

    def parse(self, data, sheet_name):
        metadata = self.parse_global_metadata(data, sheet_name)
        self.parse_ts(data, metadata)
        return self.time_series

    def data_to_string(self, data):
        return str(data)

    def parse_global_metadata(self, data, sheet_name):
        metadata = {}
        for mdname, mdspec in self.global_metadata.items():
            if mdspec['source'] == 'sheet_name':
                metadata[mdname] = sheet_name
            elif mdspec['source'] == 'cell':
                metadata[mdname] = data[mdspec['row']][mdspec['col']]
            elif mdspec['source'] == 'const':
                metadata[mdname] = mdspec['val']
            else:
                logging.warn("Unknown metadata source %s", mdspec['source'])
        metadata['provenance'] = self.provenance
        return metadata

    def parse_tsr_metadata(self, metadata, data, tsidx):
        mds = self.metadata_spec
        md_modes = {}
        all_blank = True
        for md_name in mds:
            if mds[md_name]['mode'] == 'normal' or mds[md_name]['mode'] == 'backfill':
                if mds[md_name]['source'] == 'cell':
                    metadata[md_name] = data[mds[md_name]['loc'][0]][mds[md_name]['loc'][1]]
                    if not self.is_blank(metadata[md_name]):
                        all_blank = False
                elif mds[md_name]['source'] == 'const':
                    metadata[md_name] = mds[md_name]['val']
                else:
                    md_vals = []
                    for idx in mds[md_name]['loc']:
                        coords = self.orient_coords(tsidx, idx)
                        val = self.data_to_string(data[coords[0]][coords[1]])
                        md_vals.append(val)
                        if not self.is_blank(val):
                            all_blank = False
                    metadata[md_name] = " ".join(md_vals)
                if not metadata[md_name] and mds[md_name]['mode']=='backfill':
                    metadata[md_name] = self.time_series[-1]['metadata'][md_name]
            else:
                md_modes[mds[md_name]['mode']] = True
        if all_blank and ("inline" not in md_modes or not md_modes["inline"]):
            logging.debug("%s",metadata)
            raise IndexError("All metadata values blank for %d"%(tsidx))
        return md_modes

    def parse_inline_tsr_metadata(self, metadata, data, dataidx):
        mds = self.metadata_spec
        for md_name in mds:
            if mds[md_name]['mode'] == 'inline':
                md_vals = []
                for idx in mds[md_name]['loc']:
                    coords = self.orient_coords(idx, dataidx)
                    md_vals.append(self.data_to_string(data[coords[0]][coords[1]]))
                metadata[md_name] = " ".join(md_vals)

    def orient_coords(self, tsidx, dataidx):
        if self.orientation == 'row':
            return (tsidx, dataidx)
        else:
            return (dataidx, tsidx)

    def get_uid(self, metadata):
        md_str = json.dumps(metadata, sort_keys=True, cls=DecimalJSONEncoder).encode('utf-8')
        hash_object = hashlib.sha1(md_str)
        return hash_object.hexdigest()

    def generate_time_label(self, data, d_idx):
        time_labels = []
        for tc in self.time_coordinates['locs']:
            coords = self.orient_coords(tc, d_idx)
            val = self.data_to_string(data[coords[0]][coords[1]])
            if self.is_blank(val) and self.time_coordinates['mode'] == 'backfill':
                t_idx = d_idx - 1
                while t_idx > 0 and self.is_blank(val):
                    coords = self.orient_coords(tc, t_idx)
                    val = self.data_to_string(data[coords[0]][coords[1]])
                    t_idx -= 1
            time_labels.append(val)
        time_label = " ".join(time_labels)
        if self.time_coordinates['post_process']:
            func = eval('lambda v: ' + self.time_coordinates['post_process'])
            time_label = func(time_label)
        return self.process_time_span(time_label, self.time_coordinates['granularity'])

    def process_time_span(self, time_instant, granularity):
        # TODO: other granularities added (weekly :-?)
        granularities = {'yearly', 'monthly', 'quarterly'}
        if granularity not in granularities:
            return {'instant': time_instant}

        # TODO: other parsing date patterns to be added

        date_parts = time_instant.split('-')
        if granularity == 'yearly':
            time_span = {'start_time': self.fill_date_pattern(self.parse_yearly_date(time_instant))}
            time_span['end_time'] = self.parse_yearly_date(self.parse_yearly_date(date_parts[0]))#str(int(date_parts[0]))
            time_span['end_time'] += '-' + date_parts[1] if len(date_parts) > 1 else '-12'
            time_span['end_time'] += '-' + date_parts[2] if len(date_parts) > 2 else '-30'
            return {'span': time_span}
        if granularity == 'monthly':
            time_span = {'start_time': self.fill_date_pattern(self.parse_monthly_date(time_instant))}
        else:
            time_span = {'start_time': self.fill_date_pattern(time_instant)}
        month_offset = 3
        if granularity == 'monthly':  # the final choice of time span will be based on the given day
            month_offset = 1

        month_str = str(int(date_parts[1]) + month_offset)
        year_offset = 0
        if int(date_parts[1]) + month_offset >= 13:
            month_str = str((int(date_parts[1]) + month_offset) % 12 + 1)
            year_offset += 1
            time_span['end_time'] = str(int(date_parts[0]) + 1) + '-' + str(
                (int(date_parts[1]) + month_offset) % 12 + 1)

        if len(month_str) < 2:
            month_str = '0' + month_str
        time_span['end_time'] = date_parts[0] + '-' + month_str
        time_span['end_time'] += '-' + date_parts[2] if len(date_parts) > 2 else '-01'
        return {'span': time_span}


    def parse_yearly_date(self, time_label):
        try: #TODO: add unforseen date patterns here
            return re.search('(\d{4})', time_label)[0]
        except:
            logging.error('This date format is not recognized, yearly granularity reported {}'.format(time_label))
            return '2000' # only for preventing the extractor to crash


    def parse_monthly_date(self, time_label):
        month_kw = {"jan":1, "feb":2, "mar":3, "apr":4, "may":5, "jun":6, "jul":7, "aug":8, "sep":9, "oct":10, "nov":11, "dec":12,
            "january":1, "february":2, "march":3, "april":4, "june":6, "july":7, "august":8, "september":9, "october":10, "november":11, "december":12}
        try:
            year = re.search('(\d{4})', time_label).group(1)
            month = re.search('(^|[^\d])(\d{2})([^\d]|$)', time_label)
            if month != None:
                return year + '-' + month.group(2)
            for k in month_kw:
                if k in time_label.lower:
                    return year + '-' + str(month_kw[k])
            #TODO: add unforseen date patterns here
        except:
            logging.error('This date format is not recognized, monthly granularity reported {}'.format(time_label))
        logging.error('This date format is not recognized, monthly granularity reported {}'.format(time_label))
        return '1'



    def fill_date_pattern(self, time_instant):
        date_parts = time_instant.split('-')
        time_ins = str(date_parts[0])
        time_ins += '-' + date_parts[1] if len(date_parts) > 1 else '-01'
        time_ins += '-' + date_parts[2] if len(date_parts) > 2 else '-01'
        return time_ins

    def parse_ts(self, data, metadata):
        self.time_series = []
        for ts_idx in self.series_range:
            logging.debug("Parsing TS at %d", ts_idx)
            timeseries = []
            ts_metadata = copy.deepcopy(metadata)
            ts_metadata['provenance'][self.orientation] = ts_idx
            ts_metadata['granularity']=self.time_coordinates['granularity']

            try:
                md_modes = self.parse_tsr_metadata(ts_metadata, data, ts_idx)
            except IndexError as ie:
                if type(self.series_range.curr_component()) is lr.LocationRangeInfiniteIntervalComponent:
                    logging.info("all blank metadata cells in infinite interval")
                    break
                else:
                    logging.error("metadata specifcation indexing error for time series index {}".format(ts_idx))
                    raise ie

            inline_md_curr = dict()
            inline_md_prev = None
            for d_idx in self.data_range:
                measurement = dict()
                time_label = ''
                try:
                    time_label = self.generate_time_label(data, d_idx)
                except IndexError as ie:
                    if type(self.data_range.curr_component()) is lr.LocationRangeInfiniteIntervalComponent:
                        break
                    else:
                        logging.error("metadata specifcation indexing error for data point index {}".format(d_idx))
                        raise ie

                if type(
                        self.data_range.curr_component()) is lr.LocationRangeInfiniteIntervalComponent and self.is_blank(
                        time_label):
                    logging.info("blank cell in infinite interval")
                    break

                # if inline metadata has changed (in auto-detect mode)
                # merge previous metadata
                # output old time series
                # re-initialize time series array
                if 'inline' in md_modes:
                    self.parse_inline_tsr_metadata(inline_md_curr, data, d_idx)
                    if inline_md_prev:
                        md_changed = False
                        for md_name in inline_md_prev:
                            if inline_md_curr[md_name] != inline_md_prev[md_name]:
                                md_changed = True

                        if md_changed:
                            new_metadata = dict(ts_metadata)
                            for md_name in inline_md_prev:
                                new_metadata[md_name] = inline_md_prev[md_name]

                            self.time_series.append({
                                'metadata': new_metadata,
                                'ts': timeseries
                            })
                            timeseries = []

                        inline_md_prev = inline_md_curr
                        inline_md_curr = dict()

                    else:
                        inline_md_prev = inline_md_curr
                        inline_md_curr = dict()

                coords = self.orient_coords(ts_idx, d_idx)
                measurement['value'] = data[coords[0]][coords[1]]
                measurement['time'] = time_label
                measurement['provenance'] = copy.deepcopy(ts_metadata['provenance'])
                measurement['provenance']['row'] = coords[0]
                measurement['provenance']['col'] = coords[1]
                measurement['uid'] = self.get_uid(measurement)
                timeseries.append(measurement)
            ts_metadata['uid'] = self.get_uid(ts_metadata)
            self.time_series.append(dict(metadata=ts_metadata, ts=timeseries))

    def is_blank(self, data):
        return len(str(data).strip()) == 0




import unittest
from rdflib.namespace import XSD
from etk.ontology_api import Ontology, OntologyEntity, OntologyClass
from etk.ontology_api import OntologyProperty, OntologyDatatypeProperty, OntologyObjectProperty
from etk.ontology_api import ValidationError
from etk.ontology_namespacemanager import DIG, SCHEMA


rdf_prefix = '''@prefix : <http://dig.isi.edu/ontologies/dig/> .
@prefix dc: <http://purl.org/dc/elements/1.1/> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix crm: <http://www.cidoc-crm.org/cidoc-crm/> .
@prefix schema: <http://schema.org/> .
'''

class TestOntologyAPI(unittest.TestCase):
    def test_basic_entity_info(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class ;
    rdfs:label "Entity" ;
    skos:definition """Everything that can be described, similar to owl:Thing.""" ;
    :common_properties :label, :title, :description ;
    .
        '''
        ontology = Ontology(rdf_content)
        self.assertEqual(len(ontology.all_classes()), 1)
        for entity in ontology.all_classes():
            self.assertIsInstance(entity, OntologyEntity)
            self.assertIsInstance(entity, OntologyClass)
        self.assertEqual(len(ontology.all_properties()), 0)
        for entity in ontology.all_properties():
            self.assertIsInstance(entity, OntologyEntity)
            self.assertIsInstance(entity, OntologyProperty)
        self.assertEqual(len(ontology.root_classes()), 1)
        for entity in ontology.root_classes():
            self.assertIsInstance(entity, OntologyClass)
            self.assertEqual(len(entity.super_classes()), 0)

    def test_property_domain(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class .
:Type a owl:Class .
:Place a owl:Class .
:Other a owl:Class ;
    :common_properties :code .
:code a owl:DatatypeProperty ;
    rdfs:domain [
        a owl:Class ;
        owl:unionOf (
            :Type
            :Place
        )
    ] ;
    schema:domainIncludes :Entity .
        '''
        ontology = Ontology(rdf_content)
        property_code = ontology.get_entity(DIG.code.toPython())
        self.assertIsInstance(property_code, OntologyDatatypeProperty)
        self.assertEqual(len(property_code.included_domains()), 4)
        self.assertSetEqual(property_code.included_domains(),
                            set(map(ontology.get_entity, map(str, (DIG.Entity, DIG.Type, DIG.Other, DIG.Place)))))

    def test_datatype_property_range(self):
        rdf_content = rdf_prefix + '''
:code a owl:DatatypeProperty ;
    schema:rangeIncludes xsd:string ;
    .
:iso_code a owl:DatatypeProperty ;
    rdfs:subPropertyOf :code ;
    .
        '''
        ontology = Ontology(rdf_content)
        property_code = ontology.get_entity(DIG.code.toPython())
        property_iso_code = ontology.get_entity(DIG.code.toPython())
        self.assertIsInstance(property_iso_code, OntologyDatatypeProperty)
        self.assertTrue(property_code.is_legal_object(XSD.string.toPython()))
        self.assertTrue(property_iso_code.is_legal_object(XSD.string.toPython()))
        self.assertFalse(property_code.is_legal_object(XSD.dateTime.toPython()))
        self.assertFalse(property_iso_code.is_legal_object(XSD.dateTime.toPython()))

    def test_inverse_object_property(self):
        rdf_content = rdf_prefix + '''
:moved_to a owl:ObjectProperty, rdf:Property ;
    :inverse :was_destination_of .
        '''
        ontology = Ontology(rdf_content)
        self.assertEqual(len(ontology.all_properties()), 2)
        property_moved = ontology.get_entity(DIG.moved_to.toPython())
        property_inverse = ontology.get_entity(DIG.was_destination_of.toPython())
        self.assertIsNotNone(property_moved)
        self.assertIsNotNone(property_inverse)
        self.assertTrue(property_moved.is_primary())
        self.assertFalse(property_inverse.is_primary())
        self.assertEqual(property_moved.inverse(), property_inverse)
        self.assertEqual(property_inverse.inverse(), property_moved)

    def test_class_hierachy(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class ; .
:Group a owl:Class ;
    rdfs:subClassOf :Entity ; .
:Actor a owl:Class ;
    owl:subClassOf :Group ; .
        '''
        ontology = Ontology(rdf_content)
        self.assertEqual(len(ontology.all_classes()), 3)
        entity = ontology.get_entity(DIG.Entity.toPython())
        self.assertIsNotNone(entity)
        self.assertSetEqual(ontology.root_classes(), {entity})
        group = ontology.get_entity(DIG.Group.toPython())
        self.assertIsNotNone(group)
        actor = ontology.get_entity(DIG.Actor.toPython())
        self.assertIsNotNone(actor)
        self.assertSetEqual(group.super_classes(), {entity})
        self.assertSetEqual(actor.super_classes(), {group})
        self.assertSetEqual(actor.super_classes_closure(), {entity, group})

    def test_property_hierachy(self):
        rdf_content = rdf_prefix + '''
:code a owl:DatatypeProperty ; .

:iso_code a owl:DatatypeProperty ;
    rdfs:subPropertyOf :code ; .

:cameo_code a owl:DatatypeProperty ;
    rdfs:subPropertyOf :code ; .

:iso_country_code a owl:DatatypeProperty ;
    rdfs:subPropertyOf :iso_code ; .

:adm1_code a owl:DatatypeProperty ;
    rdfs:subPropertyOf :iso_code ; .
        '''
        ontology = Ontology(rdf_content)
        self.assertEqual(len(ontology.all_classes()), 0)
        self.assertEqual(len(ontology.all_properties()), 5)
        code = ontology.get_entity(DIG.code.toPython())
        self.assertIsNotNone(code)
        self.assertIsInstance(code, OntologyDatatypeProperty)
        iso_code = ontology.get_entity(DIG.iso_code.toPython())
        self.assertIsNotNone(iso_code)
        iso_country_code = ontology.get_entity(DIG.iso_country_code.toPython())
        self.assertIsNotNone(iso_country_code)
        self.assertSetEqual(iso_code.super_properties(), {code})
        self.assertSetEqual(iso_country_code.super_properties_closure(), {code, iso_code})

    def test_class_cycle_detection(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class ;
    rdfs:label "Entity" ;
    skos:definition """Everything that can be described, similar to owl:Thing.""" ;
    owl:subClassOf :Actor ;
    :common_properties :label, :title, :description ;
    :crm_equivalent crm:E1_CRM_Entity
    .

:Group a owl:Class ;
    rdfs:label "Group" ;
    skos:definition """A """ ;
    rdfs:subClassOf :Entity ;
    :crm_equivalent crm:E74_Group ;
    :common_properties :label, :title ;
    .

:Actor a owl:Class ;
    rdfs:label "Actor" ;
    skos:definition """A group, organization, person who have the potential to do actions.""" ;
    rdfs:subClassOf :Group ;
    :crm_equivalent crm:E39_Actor ;
    :common_properties :label, :title, :religion ;
    .
        '''
        with self.assertRaises(ValidationError):
            Ontology(rdf_content)

    def test_property_cycle_detection(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class ; .
:code a owl:DatatypeProperty ;
    rdfs:label "code" ;
    skos:definition """A short string used as a code for an entity.""" ;
    skos:note """Examples are country codes, currency codes.""" ;
    schema:domainIncludes :Entity ;
    schema:rangeIncludes xsd:string ;
    rdfs:subPropertyOf :iso_code ;
    .

:iso_code a owl:DatatypeProperty ;
    rdfs:label "ISO code" ;
    skos:definition """A sanctioned ISO code for something.""" ;
    rdfs:subPropertyOf :code ;
    .
        '''
        with self.assertRaises(ValidationError):
            Ontology(rdf_content)

    def test_property_class_consistency(self):
        rdf_content = rdf_prefix + '''
:Entity a owl:Class ; .
:Place a owl:Class ; .
:code a owl:DatatypeProperty ;
    schema:domainIncludes :Entity ;
    schema:rangeIncludes xsd:string ; .

:country a owl:ObjectProperty ;
    rdfs:subPropertyOf :code ;
    schema:domainIncludes :Place ;
    schema:rangeIncludes :Country ; .
        '''
        with self.assertRaises(ValidationError):
            Ontology(rdf_content)

    def test_property_domain_inherit_consistency(self):
        rdf_content = rdf_prefix + '''
:Thing a owl:Class ; .
:Event a owl:Class ;
    rdfs:subClassOf :Thing ; .

:had_participant a owl:ObjectProperty ;
    schema:domainIncludes :Event ; .

:transferred_ownership_to a owl:ObjectProperty ;
    schema:domainIncludes :Thing ;
    owl:subPropertyOf :carried_out_by ; .

:carried_out_by a owl:ObjectProperty ;
    owl:subPropertyOf :had_participant ; .
        '''
        with self.assertRaises(ValidationError):
            Ontology(rdf_content)

    def test_property_domain_inherit_redundant(self):
        rdf_content = rdf_prefix + '''
:Event a owl:Class ; .
:had_participant a owl:ObjectProperty ;
    schema:domainIncludes :Event ;
    .
:transferred_ownership_to a owl:ObjectProperty ;
    schema:domainIncludes :Event ;
    owl:subPropertyOf :carried_out_by ;
    .
:carried_out_by a owl:ObjectProperty ;
    owl:subPropertyOf :had_participant ;
    .
        '''
        with self.assertLogs(level="WARNING"):
            Ontology(rdf_content)

    def test_property_range_inherit_consistency(self):
        rdf_content = rdf_prefix + '''
:Event a owl:Class ; .
:Entity a owl:Class ;
    .
:Group a owl:Class ;
    rdfs:subClassOf :Entity ;
    .
:Actor a owl:Class ;
    rdfs:subClassOf :Group ;
    .
:Physical_Object a owl:Class ;
    rdfs:subClassOf :Entity ;
    .
:Man_Made_Thing a owl:Class ;
    rdfs:subClassOf :Entity ;
    .
:had_participant a owl:ObjectProperty ;
    schema:rangeIncludes :Actor ;
    .
:transferred_custody_of a owl:ObjectProperty ;
    schema:rangeIncludes :Physical_Object, :Man_Made_Thing ;
    owl:subPropertyOf :had_participant ;
    .
        '''
        with self.assertRaises(ValidationError):
            Ontology(rdf_content)

    def test_property_merge_master_config(self):
        rdf_content = rdf_prefix + ':had_participant a owl:ObjectProperty . '
        ontology = Ontology(rdf_content)
        d_color = 'red'
        config = ontology.merge_with_master_config('{}', {'color': d_color})
        self.assertIn('fields', config)
        fields = config['fields']
        self.assertIn('had_participant', fields)
        property_ = fields['had_participant']
        self.assertIn('color', property_)
        self.assertEqual(property_['color'], d_color)

    def test_property_merge_config_inherit(self):
        rdf_content = rdf_prefix + '''
:had_participant a owl:ObjectProperty ; .
:transferred_custody_of a owl:ObjectProperty ;
    owl:subPropertyOf :had_participant ; .
        '''
        ontology = Ontology(rdf_content)
        config = ontology.merge_with_master_config('{"fields":{"had_participant":{"color":"red"}}}')
        self.assertIn('fields', config)
        fields = config['fields']
        self.assertIn('had_participant', fields)
        property_ = fields['had_participant']
        self.assertIn('color', property_)
        self.assertEqual(property_['color'], 'red')
        self.assertIn('transferred_custody_of', fields)
        sub_property = fields['transferred_custody_of']
        self.assertIn('color', sub_property)
        self.assertEqual(property_['color'], sub_property['color'])

    def test_property_merge_config_closest_inherit(self):
        rdf_content = rdf_prefix + '''
:had_participant a owl:ObjectProperty ; .
:carried_out_by a owl:ObjectProperty ;
    owl:subPropertyOf :had_participant ; .
:custody_received_by a owl:ObjectProperty ;
    owl:subPropertyOf :carried_out_by ; .
        '''
        ontology = Ontology(rdf_content)
        config = ontology.merge_with_master_config('{"fields":{"had_participant":{"color":"red"},'
                                                   '"carried_out_by":{"color":"blue"}}}')
        self.assertTrue('fields' in config)
        fields = config['fields']
        self.assertIn('had_participant', fields)
        property_ = fields['had_participant']
        self.assertIn('color', property_)
        self.assertEqual('red', property_['color'])
        self.assertIn('carried_out_by', fields)
        sub_property = fields['carried_out_by']
        self.assertIn('color', sub_property)
        self.assertEqual('blue', sub_property['color'])
        self.assertIn('custody_received_by', fields)
        subsub = fields['custody_received_by']
        self.assertIn('color', subsub)
        self.assertEqual(sub_property['color'], subsub['color'])

    def test_property_merge_config_delete_orphan(self):
        rdf_content = rdf_prefix + '''
:had_participant a owl:ObjectProperty ; .
:carried_out_by a owl:ObjectProperty ;
    owl:subPropertyOf :had_participant ; .
        '''
        ontology = Ontology(rdf_content)
        config = ontology.merge_with_master_config('{"fields":{"had_participant":{"color":"red"},'
                                                   '"custody_received_by":{"color":"blue"}}}',
                                                   delete_orphan_fields=True)
        self.assertTrue('fields' in config)
        fields = config['fields']
        self.assertIn('had_participant', fields)
        self.assertIn('carried_out_by', fields)
        self.assertNotIn('custody_received_by', fields)

    def test_rdf_generation(self):
        from etk.ontology_api import rdf_generation
        kg = '''
{
  "@id": "http://www.isi.edu/aida/events/dabaf6a2-744b-4f0a-a872-3c11c4aea0a9",
  "@type": ["dig:Person", "dig:Entity"],
  "label": [{
    "@value": "Jason"
  }, {
    "@value": "json"
  }],
  "@context": {
    "@vocab": "http://www.w3.org/2000/01/rdf-schema#",
    "dig": "http://dig.isi.edu/ontologies/dig/"
  }
}
        '''
        nt = rdf_generation(kg)
        self.assertIsInstance(nt, str)
        self.assertEqual(4, len([*filter(bool, nt.split('\n'))]))

    def test_ontology_api_is_valid_with_empty_kg(self):
        import json

        rdf_content = rdf_prefix + '''
:Place a owl:Class ;
    :common_properties :region ; .
:region a owl:DatatypeProperty ;
    schema:domainIncludes :Place ;
    schema:rangeIncludes xsd:string ; .
            '''
        kg = '{}'
        ontology = Ontology(rdf_content)
        self.assertTrue(ontology.is_valid('region', 'somewhere', json.loads(kg)))
        self.assertFalse(ontology.is_valid('region', 1, json.loads(kg)))
        self.assertFalse(ontology.is_valid('region', True, json.loads(kg)))

    def test_ontology_schema_org(self):
        import json

        rdf_content = rdf_prefix + '''
        schema:Person a owl:Class ; .
        schema:relatedTo
              a rdf:Property ;
              schema:domainIncludes schema:Person ;
              schema:rangeIncludes schema:Person .
        '''
        kg = '''
        {
          "@type": ["http://schema.org/Person"],
          "@id": "person1",
          "http://schema.org/relatedTo": {
            "@type": ["http://schema.org/Person"],
            "@id": "person2"
          }
        }
        '''

        ontology = Ontology(rdf_content)
        for entity in ontology.all_classes():
            self.assertIsInstance(entity, OntologyEntity)
            self.assertIsInstance(entity, OntologyClass)
        self.assertEqual(len(ontology.all_properties()), 1)

        kg_json = json.loads(kg)
        config = ontology.merge_with_master_config(kg_json,
                                                   delete_orphan_fields=True)
        self.assertTrue('fields' in config)
        fields = config['fields']
        self.assertIn('relatedTo', fields)
        #
        self.assertTrue(ontology.is_valid('schema:relatedTo',
                                          kg_json["http://schema.org/relatedTo"],
                                          kg_json))

    def test_ontology_schema_org_datatype(self):
        import json

        rdf_content = rdf_prefix + '''
        schema:Person a owl:Class ; .
        schema:Text a schema:DataType, rdfs:Class ; .
        schema:name
              a rdf:Property ;
              schema:domainIncludes schema:Person ;
              schema:rangeIncludes schema:Person .
        '''
        kg = '''
        {
          "@type": ["http://schema.org/Person"],
          "@id": "person1",
          "http://schema.org/name": "Person Name"
        }
        '''

        ontology = Ontology(rdf_content)
        for entity in ontology.all_classes():
            self.assertIsInstance(entity, OntologyEntity)
            self.assertIsInstance(entity, OntologyClass)
        self.assertEqual(len(ontology.all_properties()), 1)

        return_value = ontology.is_valid('schema:name',
                                          "person name1",
                                          json.loads(kg))
        self.assertEqual(return_value, {'@id': 'person name1'})

        rdf_content = rdf_prefix + '''
                schema:Person a owl:Class ; .
                schema:DataType a owl:Class ; .
                schema:Text a schema:DataType, owl:Class ; .
                schema:name
                      a rdf:Property ;
                      schema:domainIncludes schema:Person ;
                      schema:rangeIncludes schema:Text .
                '''
        ontology = Ontology(rdf_content)
        kg = '''
                {
                  "@type": ["http://schema.org/Person"],
                  "@id": "person1",
                  "http://schema.org/name": "Person Name"
                }
                '''
        return_value = ontology.is_valid('schema:name',
                                         "person name1",
                                         json.loads(kg))
        self.assertEqual(return_value, {'@value': 'person name1'})

    def test_ontology_condensed_json(self):
        import json

        rdf_content = rdf_prefix + '''
                schema:Person a owl:Class ; .
                schema:DataType a owl:Class ; .
                schema:Text a schema:DataType, owl:Class ; .
                schema:name
                      a rdf:Property ;
                      schema:domainIncludes schema:Person ;
                      schema:rangeIncludes schema:Text .
                '''
        ontology = Ontology(rdf_content, expanded_jsonld=False)
        kg = '''
                {
                  "@type": ["http://schema.org/Person"],
                  "@id": "person1",
                  "http://schema.org/name": "Person Name"
                }
                '''
        return_value = ontology.is_valid('schema:name',
                                         "person name1",
                                         json.loads(kg))
        self.assertEqual(return_value, 'person name1')


import logging
from typing import Set, Union, Optional
from functools import reduce
from datetime import datetime, time, date
from rdflib import Graph, URIRef, BNode
from rdflib.namespace import RDF, RDFS, OWL, SKOS, XSD
from etk.ontology_namespacemanager import OntologyNamespaceManager, SCHEMA
from itertools import chain


class OntologyEntity(object):
    """
    Superclass of all ontology objects, including classes and properties.
    """

    def __init__(self, uri):
        self._uri = str(uri)
        self._label = set()
        self._definition = set()
        self._note = set()
        self._comment = set()
        # extract name according to the definition of name
        ind = self._uri.rfind('#')
        if ind == -1:
            ind = self._uri.rfind('/')
        self._name = self._uri[ind+1:]

    def uri(self) -> str:
        """
        Returns: Tte full URI of an ontology entity.

        """
        return self._uri

    def name(self) -> str:
        """
        The name of an ontology entity is the last part of the URI after the last / or #.

        Returns: the name part of an ontology entity.

        """
        return self._name

    def label(self) -> Set[str]:
        """

        Returns: the set of rdfs:label values

        """
        return self._label

    def definition(self) -> Set[str]:
        """

        Returns: the set of skos:definition values

        """
        return self._definition

    def note(self) -> Set[str]:
        """

        Returns: the set of skos:note values

        """
        return self._note

    def comment(self) -> Set[str]:
        return self._comment

    def __str__(self):
        s = "<{}>".format(self._uri)
        for label in self._label:
            s += '\n\trdfs:label "{}";'.format(label)
        for d in self._definition:
            s += '\n\tskos:definition "{}"'.format(d)
        for n in self._note:
            s += '\n\tskos:note "{}"'.format(n)
        return s

    def __hash__(self):
        return hash(str(self._uri))

    def __eq__(self, other):
        return self._uri == other.uri()


class OntologyClass(OntologyEntity):
    def __init__(self, uri):
        super().__init__(uri)
        self._super_classes = set()

    def super_classes(self) -> Set['OntologyClass']:
        """
        super_classes(x) = the set of y_i such that x owl:subClassOf y_i.

        Returns: the set of super-classes.

        """
        return self._super_classes

    def super_classes_closure(self) -> Set['OntologyClass']:
        """

        Returns: the transitive closure of the super_classes function

        """
        if not self._super_classes:
            return set()
        return self._super_classes | reduce(set.union, [x.super_classes_closure() for x in self._super_classes])


class OntologyProperty(OntologyEntity):
    def __init__(self, uri):
        super().__init__(uri)
        self.domains = set()
        self.ranges = set()
        self._super_properties = set()

    def super_properties(self) -> Set['OntologyProperty']:
        """
        super_properties = the set of y_i such that x owl:subPropertyOf y_i.

        Returns: the set of super_properties

        """
        return self._super_properties

    def super_properties_closure(self) -> Union[Set['OntologyProperty'], Set[str]]:
        """

        Returns:  the transitive closure of the super_properties function

        """
        if not self._super_properties:
            return set()
        return self._super_properties | reduce(set.union, [x.super_properties_closure()
                                                           for x in self._super_properties])

    def included_domains(self) -> Set[OntologyClass]:
        """
        included_domains is the union of two sets:
        - the values of rdfs:domain
        - the values of schema:domainIncludes

        Although technically wrong, this function unions the rdfs:domain instead of intersecting them.
        A validation function prints a warning when loading the ontology.

        Returns: the set of all domains defined for a property.
        """
        return self.domains

    def included_ranges(self) -> Set[OntologyClass]:
        """
        included_ranges is the union of two sets:
        - the values of rdfs:range
        - the values of schema:rangeIncludes

        Although technically wrong, this function unions the rdfs:range instead of intersecting them.
        A validation function prints a warning when loading the ontology.

        Returns: the set of all ranges defined for a property.

        """
        return self.ranges

    def is_legal_subject(self, c: OntologyClass) -> bool:
        """
        is_legal_subject(c) = true if
        - c in included_domains(self) or
        - super_classes_closure(c) intersection included_domains(self) is not empty

        There is no need to check the included_domains(super_properties_closure(self)) because
        included_domains(super_properties_closure(self)) is subset of super_classes_closure(included_domains(self))

        Args:
            c:

        Returns:

        """
        domains = self.included_domains()
        return c and (not domains or c in domains or c.super_classes_closure() & domains)

    def is_legal_object(self, object) -> bool:
        raise NotImplementedError('Subclass should implement this.')


class OntologyObjectProperty(OntologyProperty):
    def __init__(self, uri, inverse_property=None):
        super().__init__(uri)
        self._inverse = None
        self._primary = True
        if inverse_property:
            self._inverse = OntologyObjectProperty(inverse_property)
            self._inverse._inverse = self
            self._inverse._primary = False

    def is_legal_object(self, c: OntologyClass) -> bool:
        """
        is_legal_object(c) = true if
        - c in included_ranges(self) or
        - super_classes_closure(c) intersection included_ranges(self) is not empty

        Args:
            c:

        Returns:

        """
        ranges = self.included_ranges()
        return not ranges or c in ranges or c.super_classes_closure() & ranges

    def inverse(self) -> 'OntologyObjectProperty':
        """
        We support a compact definition as follows:
            :moved_to a owl:ObjectProperty ;
                rdfs:label "moved to" ;
                :inverse :was_destination_of ;

        When :inverse is present, our API automatically creates an OntologyObjectProperty for it, and links
        the two OntologyObjectProperty objects as inverses of each other.

        The OntologyObjectProperty that contained the :inverse definition is marked as is_primary()

        Returns: Inverse of the object property

        """
        return self._inverse

    def is_primary(self) -> bool:
        """

        Returns: True if this object property is the primary OntologyObjectProperty, see comments in inverse()

        """
        return self._primary


class OntologyDatatypeProperty(OntologyProperty):

    def is_legal_object(self, data_type: str) -> bool:
        """
        Do data_type validation according to the rules of the XML xsd schema.

        Args:
            data_type:

        Returns:

        """
        data_type = str(data_type)
        ranges = self.included_ranges()
        return not ranges or data_type in ranges or self.super_properties() and \
               any(x.is_legal_object(data_type) for x in self.super_properties())


class ValidationError(Exception):
    """ Base class for all validation errors """
    pass


class UnsupportOntologyExpression(Exception):
    """ Base class for all unsupport Ontology expressions """
    def __init__(self, msg):
        self.message = msg

    def __str__(self):
        return "Doesn't support expression %s" % self.message


class Ontology(object):
    def __init__(self, turtle, validation=True, include_undefined_class=False, quiet=False, expanded_jsonld=True) -> None:
        """
        Read the ontology from a string containing RDF in turtle format.

        Should do ontology validation:
        - if p.sub_property_of(q) then
          - for every x in included_domains(p) we have x is subclass of some d in
        included_domains(q)
          - for every y in included_ranges(p) we have y is subclass of some r in
        included_ranges(q) -- for object properties
        - DataType properties cannot be sub-property of Object properties, and vice versa.
        - Detect cycles in sub-class and sub-property hierarchies.

        Args:
            turtle: str or Iterable[str]
        """
        import io, sys

        self.entities = dict()
        self.classes = set()
        self.object_properties = set()
        self.data_properties = set()
        self.log_stream = io.StringIO()
        self.expanded_jsonld = expanded_jsonld

        logging.getLogger().addHandler(logging.StreamHandler(self.log_stream))
        if not quiet:
            logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))

        self.g = self.__init_graph_parse(turtle)

        # Class
        for uri in self.g.subjects(RDF.type, OWL.Class):
            self.__init_ontology_class(uri)
        # Datatype Property
        for uri in self.g.subjects(RDF.type, OWL.DatatypeProperty):
            self.__init_ontology_datatype_property(uri)

        # Object Property
        for uri, inv in self.g.query("""SELECT ?uri ?inv
                                   WHERE { ?uri a owl:ObjectProperty .
                                           OPTIONAL {?uri dig:inverse ?inv }}"""):
            self.__init_ontology_object_property(uri, inv)
        for uri, inv in self.g.query("""SELECT ?uri ?inv
                                   WHERE { ?uri a rdf:Property .
                                           OPTIONAL {?uri dig:inverse ?inv }}"""):
            has_literal, all_literal = self.__check_range_is_literal(uri)
            if has_literal or all_literal:
                self.__init_ontology_datatype_property(uri)
            if not all_literal or (not has_literal and all_literal):
                self.__init_ontology_object_property(uri, inv)

        #Datatype property as range is a schema:DataType
        # for uri, range in self.g.query("""
        #                         SELECT ?uri ?range
        #                         WHERE {?uri schema:rangeIncludes ?range .
        #                                ?range a schema:DataType
        #                         }
        #                         """):
        #     self.__init_ontology_datatype_property(uri)

        # subClassOf
        for uri, sub in self.g.query("""SELECT ?uri ?sub
                                    WHERE  {{ ?uri rdfs:subClassOf ?sub }
                                              UNION { ?uri owl:subClassOf ?sub }}"""):
            self.__init_ontology_subClassOf(uri, sub, include_undefined_class)

        # subPropertyOf
        for uri, sub in self.g.query("""SELECT ?uri ?sub
                                   WHERE  {{ ?uri rdfs:subPropertyOf ?sub }
                                           UNION { ?uri owl:subPropertyOf ?sub }}"""):
            self.__init_ontology_subPropertyOf(uri, sub)

        # domain
        for uri, d in self.g.query("""SELECT ?uri ?domain
                             WHERE {{ ?uri rdfs:domain ?domain }
                                    UNION { ?uri schema:domainIncludes ?domain }
                                    UNION { ?domain dig:common_properties ?uri}}"""):
            if isinstance(d, BNode):
                for d_ in self.__read_owl_union_of(d):
                    self.__init_ontology_property_domain(uri, d_, include_undefined_class)
            else:
                self.__init_ontology_property_domain(uri, d, include_undefined_class)

        # range
        for uri, r in self.g.query("""SELECT ?uri ?range
                             WHERE {{ ?uri rdfs:range ?range }
                                    UNION { ?uri schema:rangeIncludes ?range }}"""):
            if isinstance(r, BNode):
                for r_ in self.__read_owl_union_of(r):
                    self.__init_ontology_property_range(uri, r_, include_undefined_class)
            else:
                self.__init_ontology_property_range(uri, r, include_undefined_class)

        for uri, range in self.g.query("""
                                        SELECT ?uri ?range
                                        WHERE {?uri schema:rangeIncludes ?range .
                                               ?range a schema:DataType
                                        }
                                        """):
            self.__init_ontology_property_range(uri, URIRef("http://schema.org/DataType"), include_undefined_class)
        for uri, range in self.g.query("""
                                        SELECT ?uri ?range
                                        WHERE {?uri schema:rangeIncludes ?range .
                                               ?range rdfs:subClassOf ?x .
                                               ?x a schema:DataType
                                        }
                                        """):
            self.__init_ontology_property_range(uri, URIRef("http://schema.org/DataType"), include_undefined_class)

        for entity in self.entities.values():
            uri = URIRef(entity.uri())
            for label in self.g.objects(uri, RDFS.label):
                entity._label.add(label.toPython())
            for definition in self.g.objects(uri, SKOS.definition):
                entity._definition.add(definition.toPython())
            for note in self.g.objects(uri, OWL.note):
                entity._note.add(note.toPython())
            for comment in self.g.objects(uri, RDFS.comment):
                entity._comment.add(comment.toPython())

        # After all hierarchies are built, perform the last validation
        # Validation 1: Inherited domain & range consistency
        if validation:
            for p in self.object_properties:
                self.__validation_property_domain(p)
                self.__validation_property_range(p)
            for p in self.data_properties:
                self.__validation_property_domain(p)

    def __read_owl_union_of(self, class_):
        for head in self.g.objects(class_, OWL.unionOf):
            # data range union
            return self.g.items(head)
        for head in self.g.objects(class_, OWL.onDatatype):
            # datatype restriction
            # [ a rdfs:Datatype ; owl:onDatatype DN ; owl:withRestrictions ()]
            return [head]
        for head in self.g.objects(class_, OWL.oneOf):
            # literal enumeration
            # [ a rdfs:Datatype ; owl:oneOf ()]
            return self.g.items(head)
        for _ in self.g.objects(class_, OWL.intersectionOf):
            # data range intersection
            raise UnsupportOntologyExpression("[ a rdfs:Datatype ; owl:intersectionOf ()]")
        for _ in self.g.objects(class_, OWL.datatypeComplementOf):
            # data range complement
            raise UnsupportOntologyExpression("[ a rdfs:Datatype ; owl:datatypeComplementOf ()]")
        return []

    def __convert_cons_to_list(self, head):
        list_ = []
        while head is not RDF.nil:
            for first in self.g.objects(head, RDF.first):
                list_.append(first)
            for rest in self.g.objects(head, RDF.rest):
                head = rest
        return list_

    def __check_range_is_literal(self, uri):
        has_literal, all_literal = False, True
        for range_ in chain(self.g.objects(uri, RDFS.range), self.g.objects(uri, SCHEMA.rangeIncludes)):
            if range_ == RDFS.Literal or range_ in self.xsd_ref:
                has_literal = True
            else:
                all_literal = False
        return has_literal, all_literal

    def __init_graph_parse(self, contents):
        nm = OntologyNamespaceManager(Graph())
        g = nm.graph
        if isinstance(contents, str):
            g.parse(data=contents, format='ttl')
        else:
            for content in contents:
                g.parse(data=content, format='ttl')
        return g

    def __init_ontology_class(self, uri):
        if isinstance(uri, URIRef):
            uri = uri.toPython()
        else:
            return
        entity = self.entities.get(uri, OntologyClass(uri))
        self.entities[uri] = entity
        self.classes.add(entity)
        return entity

    def __init_ontology_datatype_property(self, uri):
        uri = uri.toPython()
        entity = self.entities.get(uri, OntologyDatatypeProperty(uri))
        self.entities[uri] = entity
        self.data_properties.add(entity)
        return entity

    def __init_ontology_object_property(self, uri, inv):
        uri = uri.toPython()
        entity = self.entities.get(uri, OntologyObjectProperty(uri, inv))
        self.entities[uri] = entity
        self.object_properties.add(entity)
        if inv:
            inv = inv.toPython()
            self.entities[inv] = self.entities.get(inv, entity.inverse())
            self.object_properties.add(self.entities[inv])
        return entity

    def __init_ontology_subClassOf(self, uri, sub, include_undefined_class):
        entity = self.get_entity(uri.toPython())
        if not entity:
            if include_undefined_class:
                entity = self.__init_ontology_class(uri)
            else:
                logging.warning("Missing a class with uri %s.", uri)
                return
        if not isinstance(entity, OntologyClass):
            logging.warning("Subclass %s is not a class.", uri)
            return
        sub_entity = self.get_entity(sub.toPython())
        if not sub_entity:
            if include_undefined_class:
                sub_entity = self.__init_ontology_class(sub)
            else:
                logging.warning("Missing a super class of %s with uri %s.", uri, sub)
                return
        if not isinstance(sub_entity, OntologyClass):
            logging.warning("Superclass %s is not a class.", sub)
            return
        # Validation 3: Cycle detection
        if entity in sub_entity.super_classes_closure():
            raise ValidationError("Cycle detected")
        entity._super_classes.add(sub_entity)

    def __init_ontology_subPropertyOf(self, uri, sub):
        property_ = self.get_entity(uri.toPython())
        sub_property = self.get_entity(sub.toPython())
        if not property_:
            logging.warning("Missing a property with URI: %s", uri)
            return
        if not isinstance(property_, OntologyProperty):
            logging.warning("Subproperty %s is not a property.", uri)
            return
        if not sub_property:
            logging.warning("Misssing a subproperty of %s with URI %s", uri, sub)
            return
        if not isinstance(sub_property, OntologyProperty):
            logging.warning("Superproperty %s is not a property.", sub)
        # Validation 3: Cycle detection
        if property_ in sub_property.super_properties_closure():
            raise ValidationError("Cycle detected")
        # Validation 2: Class consistency
        if isinstance(property_, OntologyDatatypeProperty) != \
                isinstance(sub_property, OntologyDatatypeProperty):
            raise ValidationError("Subproperty {} should share the same class with {}"
                                  .format(sub_property.name(), property_.name()))
        property_._super_properties.add(sub_property)

    def __init_ontology_property_domain(self, uri, d, include_undefined_class):
        property_ = self.get_entity(uri.toPython())
        domain = self.get_entity(d.toPython())
        if not (property_ and isinstance(property_, OntologyProperty)):
            return
        if not domain:
            if include_undefined_class:
                domain = self.__init_ontology_class(d)
            else:
                logging.warning("Missing a domain class for :%s with uri %s.", property_.name(), d)
                return
        property_.domains.add(domain)

    def __init_ontology_property_range(self, uri, r, include_undefined_class):
        property_ = self.get_entity(uri.toPython())
        if not (property_ and isinstance(property_, OntologyProperty)):
            return
        if isinstance(property_, OntologyDatatypeProperty):
            property_.ranges.add(r.toPython())
            return
        range_ = self.get_entity(r.toPython())
        if not range_:
            if include_undefined_class:
                range_ = self.__init_ontology_class(r)
            else:
                logging.warning("Missing a range class for :%s with uri %s.", property_.name(), r)
                return
        if not isinstance(range_, OntologyClass):
            return
        property_.ranges.add(range_)

    def __validation_property_domain(self, p):
        """
        if p.sub_property_of(q) then
        for every x in included_domains(p) we have x is subclass of some d in
        included_domains(q)
        """
        for x in p.included_domains():
            superclasses = x.super_classes_closure()
            for q in p.super_properties_closure():
                if x in q.included_domains():
                    logging.warning("Redundant domain :%s for :%s.", x.name(), p.name())
                    continue
                if not any(d == x or d in superclasses for d in q.included_domains()) and not any(
                        d == x or d in superclasses for s in q.super_properties_closure() for d in
                        s.included_domains()):
                    raise ValidationError("Domain {} of property {} isn't a subclass of any domain of "
                                          "superproperty {}".format(x.name(), p.name(), q.name()))

    def __validation_property_range(self, p):
        """
        if p.sub_property_of(q) then
        for every y in included_ranges(p) we have y is subclass of some r in
        included_ranges(q) -- for object properties
        """
        for y in p.included_ranges():
            superclasses = y.super_classes_closure()
            for q in p.super_properties():
                if not any(r == y or r in superclasses for r in q.included_ranges()) and not any(
                        r == y or r in superclasses for s in q.super_properties_closure() for r in
                        s.included_ranges()):
                    raise ValidationError("Range {} of property {} isn't a subclass of any range of"
                                          " superproperty {}".format(y.name(), p.name(), q.name()))

    def all_properties(self) -> Set[OntologyProperty]:
        return self.data_properties | self.object_properties

    def all_classes(self) -> Set[OntologyClass]:
        return self.classes

    def get_entity(self, uri: str) -> OntologyClass:
        """
        Find an ontology entity based on URI

        :param uri: URIRef or str
        :return: the OntologyEntity having the specified uri, or None
        """
        return self.entities.get(str(uri), None)

    def root_classes(self) -> Set[OntologyClass]:
        """
        Returns: All classes that don't have a super class.
        """
        return set(filter(lambda e: not e.super_classes(), self.classes))

    def html_documentation(self, include_turtle=False, exclude_warning=False) -> str:
        from etk.ontology_report_generator import OntologyReportGenerator
        return OntologyReportGenerator(self).generate_html_report(include_turtle, exclude_warning)

    def merge_with_master_config(self, config, defaults={}, delete_orphan_fields=False) -> dict:
        """
        Merge current ontology with input master config.

        :param config: master config, should be str or dict
        :param defaults: a dict that sets default color and icon
        :param delete_orphan_fields: if a property doesn't exist in the ontology then delete it
        :return: merged master config in dict
        """
        if isinstance(config, str):
            import json
            config = json.loads(config)
        properties = self.all_properties()
        config['fields'] = config.get('fields', dict())
        fields = config['fields']

        d_color = defaults.get('color', 'white')
        d_icon = defaults.get('icon', 'icons:default')

        if delete_orphan_fields:
            exist = {p.name() for p in properties}
            unexist = set(fields.keys()) - exist
            for name in unexist:
                del fields[name]

        for p in properties:
            field = fields.get(p.name(), {'show_in_search': False,
                                          'combine_fields': False,
                                          'number_of_rules': 0,
                                          'glossaries': [],
                                          'use_in_network_search': False,
                                          'case_sensitive': False,
                                          'show_as_link': 'text',
                                          'blacklists': [],
                                          'show_in_result': 'no',
                                          'rule_extractor_enabled': False,
                                          'search_importance': 1,
                                          'group_name': '',
                                          'show_in_facets': False,
                                          'predefined_extractor': 'none',
                                          'rule_extraction_target': ''})
            config['fields'][p.name()] = field
            field['screen_label'] = ' '.join(p.label())
            field['description'] = '\n'.join(p.definition())
            field['name'] = p.name()

            # color
            if 'color' not in field:
                color = self.__merge_close_ancestor_color(p, fields, attr='color')
                field['color'] = color if color else d_color
            # icon
            if 'icon' not in field:
                icon = self.__merge_close_ancestor_color(p, fields, attr='icon')
                field['icon'] = icon if icon else d_icon
            # type
            if isinstance(p, OntologyObjectProperty):
                field['type'] = 'kg_id'
            else:
                try:
                    field['type'] = self.__merge_xsd_to_type(next(iter(p.included_ranges())))
                except StopIteration:
                    field['type'] = None
        return config

    def __merge_close_ancestor_color(self, property_, fields, attr):
        from collections import deque
        added = property_.super_properties()
        ancestors = deque(added)
        while ancestors:
            super_property = ancestors.popleft()
            name = super_property.name()
            if name in fields and attr in fields[name]:
                return fields[name][attr]
            ancestors.extend(list(super_property.super_properties() - added))

    def __merge_xsd_to_type(self, uri):
        return self.xsd_ref.get(URIRef(uri), None)

    xsd_ref = {
        XSD.string: 'string',
        XSD.boolean: 'number',
        XSD.decimal: 'number',
        XSD.float: 'number',
        XSD.double: 'number',
        XSD.duration: 'number',
        XSD.dateTime: 'date',
        XSD.time: 'date',
        XSD.date: 'date',
        XSD.gYearMonth: 'date',
        XSD.gYear: 'date',
        XSD.gMonthDay: 'date',
        XSD.gMonth: 'date',
        XSD.gDay: 'date',
        XSD.hexBinary: 'string',
        XSD.base64Binary: 'string',
        XSD.anyURI: 'string',
        XSD.QName: 'string',
        XSD.NOTATION: 'string'
    }

    def is_valid(self, field_name: str, value, kg: dict) -> Optional[dict]:
        """
        Check if this value is valid for the given name property according to input knowledge graph and ontology.
        If is valid, then return a dict with key @id or @value for ObjectProperty or DatatypeProperty.
        No schema checked by this function.

        :param field_name: name of the property, if prefix is omitted, then use default namespace
        :param value: the value that try to add
        :param kg: the knowledge graph that perform adding action
        :return: None if the value isn't valid for the property, otherwise return {key: value}, key is @id for
            ObjectProperty and @value for DatatypeProperty.
        """
        # property
        uri = self.__is_valid_uri_resolve(field_name, kg.get("@context"))
        property_ = self.get_entity(uri)
        if not isinstance(property_, OntologyProperty):
            logging.warning("Property is not OntologyProperty, ignoring it:  %s", uri)
            return None
        if not self.__is_valid_domain(property_, kg):
            logging.warning("Property does not have valid domain, ignoring it:  %s", uri)
            return None
        # check if is valid range
        # first determine the input value type
        if isinstance(property_, OntologyDatatypeProperty):
            types = self.__is_valid_determine_value_type(value)
        else:
            if isinstance(value, dict):
                try:
                    types = map(self.get_entity, value['@type'])
                except KeyError:
                    return None  # input entity without type
            elif self.__is_schema_org_datatype(property_):
                if self.expanded_jsonld:
                    return {'@value': self.__serialize_type(value)}
                else:
                    return value
            else:
                return {'@id': self.__serialize_type(value)}
        # check if is a valid range
        if any(property_.is_legal_object(type_) for type_ in types):
            if isinstance(property_, OntologyObjectProperty):
                return value
            elif self.expanded_jsonld:
                return {'@value': self.__serialize_type(value)}
            else:
                return self.__serialize_type(value)
        return None

    @staticmethod
    def __is_schema_org_datatype(property_):
        ranges = property_.ranges
        for range_ in ranges:
            if range_._uri == "http://schema.org/DataType" or range_._uri == "schema:DataType":
                return True
            for super_class in range_.super_classes_closure():
                if super_class._uri == "http://schema.org/DataType" or range_._uri == "schema:DataType":
                    return True
        return False

    @staticmethod
    def __serialize_type(value):
        if isinstance(value, datetime):
            return value.isoformat()
        return value

    @staticmethod
    def __is_valid_determine_value_type(value):
        type_infer = {
            int: {XSD.int, XSD.duration, XSD.boolean, XSD.gYear, XSD.gMonth, XSD.gDay},
            float: {XSD.float, XSD.decimal, XSD.double, XSD.duration},
            bool: {XSD.boolean},
            str: {XSD.string, XSD.hexBinary, XSD.base64Binary, XSD.anyURI, XSD.QName, XSD.NOTATION},
            datetime: {XSD.datetime},
            time: {XSD.time},
            date: {XSD.date}
        }
        for class_ in type_infer:
            if isinstance(value, class_):
                return [x.toPython() for x in type_infer[class_]]
        return None

    def __is_valid_domain(self, property_, kg):
        import json
        # extract id a.k.a uri from kg
        empty_kg = True
        kg_ = Graph().parse(data=json.dumps(kg), format='json-ld')
        for type_ in kg_.objects(None, RDF.type):
            empty_kg = False
            entity = self.get_entity(type_)
            if entity and property_.is_legal_subject(entity):
                return True
        return empty_kg

    def __is_valid_uri_resolve(self, field_name: str, context: Optional[dict]) -> URIRef:
        uri = OntologyNamespaceManager.check_uriref(field_name)
        if not uri and context:
            if field_name in context:
                uri = context[field_name]
            else:
                try_split = field_name.split(':')
                if len(try_split) == 2:
                    prefix, name = try_split
                    if prefix in context:
                        uri = context[prefix] + name
                elif '@vocab' in context:
                    uri = context['@vocab'] + field_name
        if not uri:
            uri = self.g.namespace_manager.parse_uri(field_name)
        return uri


def rdf_generation(kg_object) -> str:
    """
    Convert input knowledge graph object into n-triples RDF

    :param kg_object: str, dict, or json object
    :return: n-triples RDF in str
    """
    import json

    if isinstance(kg_object, dict):
        kg_object = json.dumps(kg_object)
    g = Graph()
    g.parse(data=kg_object, format='json-ld')
    return g.serialize(format='nt').decode('utf-8')



class LocationRangeComponent(object):
    def __init__(self, rtype):
        self.rtype = rtype


class LocationRangeSingletonComponent(LocationRangeComponent):
    def __init__(self, val, rtype='singleton'):
        super(LocationRangeSingletonComponent, self).__init__(rtype)
        self.val = val

    def generator(self):
        yield self.val


class LocationRangeIntervalComponent(LocationRangeComponent):
    def __init__(self, start, end, rtype='range', increment=1):
        super(LocationRangeIntervalComponent, self).__init__(rtype)
        self.start = start
        self.end = end
        self.incr = increment

    def generator(self):
        for x in range(self.start, self.end, self.incr):
            yield x


class LocationRangeInfiniteIntervalComponent(LocationRangeComponent):
    def __init__(self, start, rtype='range', increment=1):
        super(LocationRangeInfiniteIntervalComponent, self).__init__(rtype)
        self.start = start
        self.incr = increment

    def generator(self):
        x = self.start
        while True:
            yield x
            x += self.incr


class LocationRange(object):
    def __init__(self):
        self.vals = []
        self.iteridx=0
        self.generator = None

    def add_component(self, component):
        self.vals.append(component)

    def curr_component(self):
        return self.vals[self.iteridx]

    def __iter__(self):
        self.iteridx=0
        self.generator = None
        return self

    def __next__(self):
        if self.iteridx >= len(self.vals):
            raise StopIteration
        elif self.generator is None:
            self.generator=self.vals[self.iteridx].generator()

        try:
            return self.generator.__next__()
        except StopIteration:
            self.generator = None
            self.iteridx += 1
            return self.__next__()

        next = __next__



import unittest
from rdflib import Graph, URIRef
from etk.ontology_namespacemanager import OntologyNamespaceManager
from etk.ontology_namespacemanager import PrefixNotFoundException, PrefixAlreadyUsedException


class TestOntologyNamespaceManager(unittest.TestCase):
    def test_ontology_namespace_parse_uri(self):
        nm = OntologyNamespaceManager(Graph())
        uri = URIRef('http://dig.isi.edu/ontologies/dig/Event')
        self.assertEqual(uri, nm.parse_uri(uri))
        for case in ('http://dig.isi.edu/ontologies/dig/Event', 'dig:Event'):
            self.assertEqual(uri, nm.parse_uri(case))
        with self.assertRaises(PrefixNotFoundException):
            nm.parse_uri(':Event')
        nm.bind('', 'http://dig.isi.edu/ontologies/dig/')
        self.assertEqual(uri, nm.parse_uri(':Event'))
        self.assertEqual(uri, nm.parse_uri('Event'))

    def test_ontology_namespace_bind(self):
        correct_content = '@prefix : <http://w.org/> . @prefix schema: <http://schema.org/> .'
        wrong_content = '@prefix schema: <http://dig.schema.org/> .'
        nm = OntologyNamespaceManager(Graph())
        namespace = {x[0] for x in nm.namespaces()}
        self.assertNotIn('', namespace)
        for name in ('owl', 'rdfs', 'rdf', 'schema', 'xsd', 'xml', 'dig'):
            self.assertIn(name, namespace)
        with self.assertRaises(PrefixAlreadyUsedException):
            nm.graph.parse(data=wrong_content, format='ttl')
        nm.graph.parse(data=correct_content, format='ttl')
        namespace = {x[0] for x in nm.namespaces()}
        self.assertIn('', namespace)



from etk.extraction import Extractable
from typing import List, Dict
from etk.origin_record import OriginRecord


class StorageProvenanceRecord(Extractable):
    """
    An individual storage provenance record of a storage of set of a set of extraction results from a document.
    """

    def __init__(self, id, json_path: str, attribute: str, extraction_provenances,
                 _document=None) -> None:

        Extractable.__init__(self)
        self.id = id
        self.field = None
        self._destination = json_path + '.' + attribute
        self.extraction_provenances = extraction_provenances  # will be assigned later_provenances
        self._document = _document
        self.doc_id = None

    @property
    def document(self):
        """
        Returns: the parent Document
        """
        return self._document

    @property
    def destination(self):
        """
        Returns: the parent Document
        """
        return self._destination

    def get_origins(self, value):

        prov_ids = self.extraction_provenances
        if value in prov_ids:
            extraction_prov_id = prov_ids[value]
            extraction_prov = self._document.provenances[extraction_prov_id]
            origins = extraction_prov.get_origins(value)
            return origins
        else:
            originRecord = OriginRecord(self._destination, None, None, self._document)
            return [originRecord]
